extern "C" {
#include <pcap.h>
}
#ifdef WIN32
#include <Win32-Extensions.h>
#endif

/**
 * libpcap/Winpcap C++ wrapper classes.
 * Copyright Esmond Pitt, 2008. All rights reserved.
 */

// Base exception class
class PcapException
{
public:
	// Constructor
	PcapException(const char *message)
		: _message(strdup(message))
	{
		printf("PcapException: %s\n", message);
	}
	// Copy constructor
	PcapException(const PcapException &that)
		: _message(strdup(that._message))
	{
	}
	// Destructor
	~PcapException()
	{
		delete _message;
	}
	// Get the message
	char	*getMessage() const	{ return _message; }
private:
	char	*_message;
};

class PcapFilterException : public PcapException
{
public:
	PcapFilterException(const char *msg)
		: PcapException(msg)
	{
	}
};

class PcapDeviceLookupException : public PcapException
{
public:
	PcapDeviceLookupException(const char *msg)
		: PcapException(msg)
	{
	}
};

/**
 * Base class, mainly to define a virtual constructor
 */
class PcapBase
{
public:
	virtual ~PcapBase(){}
};

/**
 * Compiled filter, or BPF program.
 */
class PcapCompiledFilter : public PcapBase
{
public:
	/**
	 * Constructor.
	 * @param pc pcap_t
	 * @param filter Filter expression
	 * @param optimize Optimize flag
	 * @param netmask Netmask, in network byte order.
	 */
	PcapCompiledFilter(pcap_t *pc, const char *filter, bool optimize, int netmask)
		throw(PcapFilterException)
		: _compiled(0)
	{
		if (pcap_compile(pc, &_bpf, const_cast<char*>(filter), optimize, netmask) == -1)
			// if the filter wouldn't compile, throw an exception back to java
			throw PcapFilterException(pcap_geterr(pc));
		_compiled = 1;
	}
	virtual ~PcapCompiledFilter()
	{
		if (_compiled)
			pcap_freecode(&_bpf);
	}
	// get the filter
	bpf_program	*getFilter()
	{
		return &_bpf;
	}
private:
	bpf_program	_bpf;
	bool	_compiled;
};

/*
 * Unified class containing all JNI-side PacketCapture instance state.
 * This has made it possible to eliminate the old concept
 * of instance numbers and all the associated arrays.
 * EJP 23 May 2008.
 */
class PacketCapture : public PcapBase
{
public:
	// Constructor
	PacketCapture()
		: _pd(0),
		_going(FALSE),
		_dumper(0),
		// This is where the default loopMode is set.
		_capture(pcap_loop),
		_captureCount(0)
	{
	}

	// Destructor
	~PacketCapture()
	{
		closePD();
		closeDumper();
	}

	// Get the capture count
	int	getCaptureCount() const			{ return _captureCount;	}

	// Get the capture device
	pcap_t	*getPD() const				{ return _pd; }

	// Get the dumper
	pcap_dumper_t	*getDumper() const	{ return _dumper; }
	
	// return true iff pcap_loop()/pcap_dispatch() should continue
	bool	isGoing() const				{ return _going; }
	
	// Set the capture device
	void	setPD(pcap_t *pd)					{ _pd = pd; }
	
	// Set the dumper
	void	setDumper(pcap_dumper_t *dumper)	{ _dumper = dumper; }
	
	// Enable pcap_loop()/pcap_dispatch()
	void	go()
	{
		_going = TRUE;
	}
	
	// Stop pcap_loop()/pcap_dispatch()
	void	stop()
	{
		_going = FALSE;
	}

	// Close the capture device
	void	closePD()
	{
		if (_pd)
			pcap_close(_pd);
		_pd = 0;
	}
	
	// Close the dumper
	void	closeDumper()
	{
		if (_dumper)
			pcap_dump_close(_dumper);
		_dumper = 0;
	}

	// Execute the capture,
	// either pcap_loop or pcap_dispatch depending on the loop mode.
	int	capture(int count, pcap_handler handler, u_char *user)
	{
		int	rc = _capture(getPD(), count, handler, user);
		// NOTE: pcap_loop() returns 0 if 'count' was exhausted,
		// which needs to be translated to 'count'.
		// pcap_dispatch() returns the actual number of packets received.
		// Both return a negative value on error.
		if (_capture == pcap_loop && rc == 0 && count > 0)
			rc = count;
		_captureCount = rc;
//		printf("captured %d/%d packets with loop_function=%s\n", _captureCount, count, _capture == pcap_loop ? "pcap_loop" : "pcap_dispatch"); fflush(stdout);
		return _captureCount;
	}

	// Set the loop mode => the capture loop function.
	void	setLoopMode(bool loopMode)
	{
		if (loopMode)
			_capture = pcap_loop;
		else
			_capture = pcap_dispatch;
	}

private:
	pcap_t	*_pd;
	bool	_going;
	// Packet dump support. Esmond Pitt 9 May 2008.
	pcap_dumper_t	*_dumper;
	// EJP 26 May 2008: added support for *both* pcap_loop() and pcap_dispatch().

	int (*_capture)(pcap_t *pc, int count, pcap_handler handler, u_char *user);
	int	_captureCount;
};

/*
 * Wrapper for pcap_findalldevs()/pcap_findalldevs_ex(),
 * ensuring the pcap_if_t data is always freed on any exit.
 */
class PcapInterfaces : public PcapBase
{
public:
	PcapInterfaces()
		throw(PcapDeviceLookupException)
		: _devs(0)
	{
		char ebuf[PCAP_ERRBUF_SIZE];
		int	rc = pcap_findalldevs(&_devs, ebuf);
		if (rc < 0)
		    throw PcapDeviceLookupException(ebuf);
	}
#ifdef WIN32
	PcapInterfaces(char *source, pcap_rmtauth *auth)
		throw(PcapDeviceLookupException)
		: _devs(0)
	{
		char ebuf[PCAP_ERRBUF_SIZE];
		int rc = pcap_findalldevs_ex(source, auth, &_devs, ebuf);
		if (rc < 0)
		    throw PcapDeviceLookupException(ebuf);
	}
#endif
	// Destructor
	~PcapInterfaces()
	{
		if (_devs)
			pcap_freealldevs(_devs);
	}
	// Get the devices pointer
	pcap_if_t	*getDevices()
	{
		return _devs;
	}
private:
	pcap_if_t *_devs;
};

#ifdef WIN32
// Everything else is only present if it is supported, i.e. if it will compile!

class PcapSendQueueException : public PcapException
{
public:
	PcapSendQueueException(const char *msg)
		: PcapException(msg)
	{
	}
};

/*
 * PcapSendQueue class.
 * Note: a pcap_send_queue can't be reused after calling pcap_sendqueue_transmit.
 * This class copes with this behaviouy by creating a new queue as required.
 */
class PcapSendQueue : public PcapBase
{
public:
	PcapSendQueue(PacketCapture *out, unsigned queueLength, bool timed)
		throw(PcapSendQueueException)
		: _out(out),
		_queue(create(queueLength)),
		_queueLength(queueLength),
		_timed(timed),
		_pending(0)
	{
	}

	~PcapSendQueue() throw(PcapSendQueueException)
	{
		if (_pending > 0)
			flush();
		destroy();
	}

	static pcap_send_queue	*create(unsigned queueLength)
		throw(PcapSendQueueException)
	{
		pcap_send_queue	*queue = pcap_sendqueue_alloc(queueLength);
		if (queue == 0)
		    throw PcapSendQueueException("can't create pcap_sendqueue");
		// printf("PcapSendQueue.create(): created the queue length %u\n", queueLength); fflush(stdout);
		return queue;
	}

	void	destroy()
	{
		if (_queue != 0)
		{
			pcap_sendqueue_destroy(_queue);
			// printf("PcapSendQueue.destroy(): destroyed the queue\n"); fflush(stdout);
		}
		_queue = 0;
	}

	void	flush() throw(PcapSendQueueException)
	{
		if (_pending > 0)
		{
			if (pcap_sendqueue_transmit(_out->getPD(), _queue, _timed) < 0)
				throw PcapSendQueueException(pcap_geterr(_out->getPD()));
			// printf("PcapSendQueue.queue(): flushed %d bytes\n", _pending); fflush(stdout);
			_pending = 0;
			// The queue is now useless and must be destroyed and reallocated
			destroy();
		}
	}

	pcap_send_queue	*getQueue() throw(PcapSendQueueException)
	{
		// Lazy allocation of the queue
		if (_queue == 0)
			_queue = create(_queueLength);
		return _queue;
	}

	int	queue(const struct pcap_pkthdr *h, const u_char *data)
		throw(PcapSendQueueException)
	{
		// printf("PcapSendQueue.queue() caplen=%d len=%d pending=%d queueLength=%d\n", h->caplen, h->len, _pending, _queueLength); fflush(stdout);
		// flush() required?
//		if (_pending+h->caplen > _queueLength)
//			flush();
		// That's not a reliable test, as there must be queue overheads.
		// The best test, as always, is to try it.
		int	rc = pcap_sendqueue_queue(getQueue(), h, data);
		if (rc < 0)
		{
			// Try one flush() and repeat
			flush();
			rc = pcap_sendqueue_queue(getQueue(), h, data);
		}
		if (rc < 0)
			throw PcapSendQueueException("can't queue to pcap_send_queue");
		else
			_pending += h->caplen;
		// printf("PcapSendQueue.queue(): queued %u bytes rc=%d\n", h->caplen, rc); fflush(stdout);
		return rc;
	}

private:
	PacketCapture	*_out;
	pcap_send_queue	*_queue;
	unsigned		_queueLength;
	bool			_timed;
	int				_pending;
};
#endif
