// $Id: CaptureTest.java,v 1.3 2008/08/28 04:18:59 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Copyright (C) 2008, Esmond Pitt                                         *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.capture;
import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.net.InetAddress;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.Arrays;
import net.sourceforge.jpcap.net.ARPFields;
import net.sourceforge.jpcap.net.ARPPacket;
import net.sourceforge.jpcap.net.IPPacket;
import net.sourceforge.jpcap.net.Packet;
import net.sourceforge.jpcap.net.RawPacket;


/**
 * Class to exercise PacketCapture.
 * <p>
 * Testing of PacketCapture could easily be implemented in the same class's
 * main method, but since the tester is a client and must implement the
 * listener interface, seems clearer to decouple the two.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.3 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/28 04:18:59 $
 */
public class CaptureTest implements PacketListener, RawPacketListener
{
	protected static String NAME = "CaptureTest";
	protected static int TEST_COUNT = 25;
	protected static String	Filter = "ip";
	private File	file = new File("test_captures.cap");
	int	packetNumber, rawPacketNumber;
	
	/**
	 * Create a new tester.
	 */
	public CaptureTest()
	{
	}
	
	private boolean	arpTested,
		ipTested;
	// implementation of PacketListener interface
	public void	packetArrived(Packet packet)
	{
		System.out.println(NAME + " #"+(++packetNumber)+": " + packet + " arrived");
		if (!arpTested && packet instanceof ARPPacket)
		{
			try
			{
				ARPPacket	ap = (ARPPacket)packet;
				InetAddress	address = InetAddress.getByName("127.0.2.2");
				ap.setSourceProtoAddress(address);
				assert ap.getSourceProtoAddress().equals("127.0.2.2") : "ARPPacket.setSourceProtoAddress failed: result="+ap.getSourceProtoAddress();
				ap.setDestinationProtoAddress(address);
				assert ap.getDestinationProtoAddress().equals("127.0.2.2") : "ARPPacket.setDestinationProtoAddress failed: result="+ap.getDestinationProtoAddress()+":"+new BigInteger(1, ap.getEthernetHeader()).toString(16)+":"+new BigInteger(1, ap.getEthernetData()).toString(16);
				ap.setOperation(ARPFields.ARP_OP_REQ_CODE);
				assert ap.getOperation() == 1 : "ARPPacket.setOperation failed: result="+ap.getOperation()+":"+new BigInteger(1, ap.getEthernetHeader()).toString(16)+":"+new BigInteger(1, ap.getEthernetData()).toString(16);
				System.out.println("ARPPacket set methods OK");
				arpTested = true;
			}
			catch (UnknownHostException exc)
			{
				exc.printStackTrace();
			}
		}
		if (!ipTested && packet instanceof IPPacket)
		{
			try
			{
				IPPacket	ip = (IPPacket)packet;
				InetAddress	address = InetAddress.getByName("127.0.2.2");
				ip.setSourceInetAddress(address);
				assert ip.getSourceAddress().equals("127.0.2.2") : "IPPacket.setSourceAddress failed: result="+ip.getSourceAddress()+":"+new BigInteger(1, ip.getEthernetHeader()).toString(16)+":"+new BigInteger(1, ip.getEthernetData()).toString(16);
				ip.setDestinationInetAddress(address);
				assert ip.getDestinationAddress().equals("127.0.2.2") : "IPPacket.setDestinationAddress failed: result="+ip.getDestinationAddress()+":"+new BigInteger(1, ip.getEthernetHeader()).toString(16)+":"+new BigInteger(1, ip.getEthernetData()).toString(16);
				System.out.println("IPPacket set methods OK");
				ipTested = true;
			}
			catch (UnknownHostException exc)
			{
				exc.printStackTrace();
			}
		}
	}
	
	// implementation of RawPacketListener interface
	
	public void rawPacketArrived(RawPacket rawPacket)
	{
		System.out.println(NAME + " #"+(++rawPacketNumber)+": " + rawPacket + " arrived");
	}
	
	
	// implementation of test logic
	
	/**
	 * Execute the test.
	 */
	public void runTest()
	{
		System.out.println("Library version: "+PacketCapture.getLibVersion());
		
		//-- create PacketCapture system
		System.out.println(NAME + ": instantiating PacketCapture object.. ");
		PacketCapture pc = new PacketCapture();
		System.out.println(NAME + ": PacketCapture instantiated ok");
		
		//-- set the capture device
		String device = null;
		try
		{
			System.out.println(NAME + ": detecting capture devices: ");
			CaptureDevice[]	devs = PacketCapture.lookupCaptureDevices();
			for (CaptureDevice dev : devs)
				System.out.println("\t"+dev);
			System.out.print(NAME + ": detecting capture device.. ");
			device = pc.findDevice();
			System.out.println(device+":network="+Integer.toHexString(pc.getNetwork(device))+":netmask="+Integer.toHexString(pc.getNetmask(device)));
		}
		catch(CaptureDeviceLookupException e)
		{
			// usually occurs if no network devices are available on the system
			System.out.println(e);
			System.exit(1);
		}
		catch(CaptureDeviceNotFoundException e)
		{
			// usually occurs if no network devices are available on the system
			System.out.println(e);
			System.exit(1);
		}
		catch(CaptureConfigurationException e)
		{
			// Shouldn't happen unless the device name is wrong
			System.out.println(e);
			System.exit(1);
		}
		
		//-- open the capture device
		try
		{
			System.out.print(NAME + ": opening capture device.. ");
			pc.open(device, true);
			System.out.println(device);
		}
		catch(CaptureDeviceOpenException e)
		{
			// usually occurs if the user doesn't have sufficient privileges
			System.out.println(e);
			System.exit(1);
		}
		
		try
		{
			// Exercise link layer type API
			System.out.println("Link layer name="+pc.getLinkLayerName());
			System.out.println("Link layer description="+pc.getLinkLayerDescription());
			int	linkTypes[] = pc.getLinkLayerTypes();
			for (int linkType : linkTypes)
			{
				System.out.println("Device supports link layer type "+linkType);
				System.out.println("\tname="+pc.getLinkLayerName(linkType));
				System.out.println("\tdescription="+pc.getLinkLayerDescription(linkType));
			}
		}
		catch (CaptureConfigurationException exc)
		{
			exc.printStackTrace();
			System.exit(1);
		}
		
		//-- set filters
		try
		{
			System.out.println(NAME + ": setting filter to '" + Filter + "'.. ");
			pc.setFilter(Filter, true);
			System.out.println(NAME + ": filter compiled and activated ok");
		}
		catch(InvalidFilterException e)
		{
			System.out.println(e);
			System.exit(1);
		}
		
		//-- register the tester as a packet listener
		System.out.print(NAME + ": registering as a packet listener.. ");
		pc.addPacketListener(this);
		System.out.println("ok");
		
		//-- register the tester as a raw packet listener
		System.out.print(NAME + ": registering as a raw packet listener.. ");
		pc.addRawPacketListener(this);
		System.out.println("ok");
		
		//-- capture packets
		try
		{
			int count = TEST_COUNT;
			System.out.println(NAME + ": waiting for " + count + " packet(s).. ");
			pc.capture(count);
			System.out.println(NAME + ": done capturing.");
		}
		catch(CapturePacketException e)
		{
			e.printStackTrace();
			System.exit(1);
		}
		
		//-- dump statistics
		System.out.println(pc.getStatistics());
		
		try
		{
			System.out.println(NAME+": testing captureNext()");
			// Capture one packet directly
			Packet	p = pc.captureNext();
			System.out.println(NAME+": captured "+p);
		}
		catch (JpcapException exc)
		{
			exc.printStackTrace();
		}
		catch (SocketTimeoutException exc)
		{
			System.out.println(NAME+": captureNext: "+exc);
		}
		catch (IOException exc)
		{
			exc.printStackTrace();
		}
		
		pc.close();
		
		// Test offline functions
		try
		{
			testLookupDevice();
			testDump();
			testCaptureTo();
			testCaptureNext();
			testCaptureToNet();
			testInstances();
		}
		catch (Exception exc)
		{
			System.out.println("runTest() (offline functions): caught "+exc);
			exc.printStackTrace();
			System.exit(1);
		}
		
	}
	
	// Exercise same as above but via the findLookupDevice() method,
	// which also exercises listLookupDevices().
	void	testLookupDevice() throws Exception
	{
		
		PacketCapture pc = null;
		
		//-- set the capture device
		CaptureDevice device = null;
		try
		{
			System.out.print(NAME + ": detecting capture device.. ");
			device = PacketCapture.findCaptureDevice();
			System.out.println(device);
		}
		catch(CaptureDeviceNotFoundException e)
		{
			// usually occurs if no network devices are available on the system
			System.out.println(e);
			System.exit(1);
		}
		catch(CaptureConfigurationException e)
		{
			// Shouldn't happen unless the device name is wrong
			System.out.println(e);
			System.exit(1);
		}
		
		//-- open the capture device
		try
		{
			System.out.print(NAME + ": opening capture device.. ");
			pc = new PacketCapture(device, true);
			System.out.println(device);
		}
		catch(CaptureDeviceOpenException e)
		{
			// usually occurs if the user doesn't have sufficient privileges
			System.out.println(e);
			System.exit(1);
		}
		
		try
		{
			// Exercise link layer type API
			System.out.println("Link layer type="+pc.getLinkLayerDescription());
		}
		catch (CaptureConfigurationException exc)
		{
			exc.printStackTrace();
			System.exit(1);
		}
		
		//-- set filters
		try
		{
			String filter = ""; // nothing filtered by default
			System.out.println(NAME + ": setting filter to '" + filter + "'.. ");
			pc.setFilter(filter, true);
			System.out.println(NAME + ": filter compiled and activated ok");
		}
		catch(InvalidFilterException e)
		{
			System.out.println(e);
			System.exit(1);
		}
		
		//-- register the tester as a packet listener
		System.out.print(NAME + ": registering as a packet listener.. ");
		pc.addPacketListener(this);
		System.out.println("ok");
		
		//-- register the tester as a raw packet listener
		System.out.print(NAME + ": registering as a raw packet listener.. ");
		pc.addRawPacketListener(this);
		System.out.println("ok");
		
		//-- capture packets
		try
		{
			int count = TEST_COUNT;
			System.out.println(NAME + ": waiting for " + count + " packet(s).. in Loop mode");
			packetNumber = rawPacketNumber = 0;
			pc.capture(count);
			System.out.println(NAME + ": done capturing.");
			
			System.out.println(NAME + ": waiting for " + count + " packet(s).. in Dispatch mode");
			pc.setLoopMode(PacketCapture.LoopMode.Dispatch);
			packetNumber = rawPacketNumber = 0;
			pc.capture(count);
			System.out.println(NAME + ": done capturing.");
		}
		catch(CapturePacketException e)
		{
			e.printStackTrace();
			System.exit(1);
		}
		
		//-- dump statistics
		System.out.println(pc.getStatistics());
		
		pc.close();
	}
	
// Exercise dump functions
	public void	testDump() throws Exception
	{
		CaptureDevice	device = PacketCapture.findCaptureDevice();
		System.out.println("Device="+device);
		// Open the device.
		PacketCapture	pc = new PacketCapture(device, 1500, true, 5*1000);
		System.out.println("opened "+device);
		
		// Test output stream
		PacketOutputStream	out = pc.getOutputStream(file);
		System.out.println("Opened dump file "+file);
		
		// Test input stream
		PacketInputStream	in = pc.getInputStream();
		System.out.println("got packet input stream");
		for (int i = 1; i <= TEST_COUNT; i++)
		{
			try
			{
				Packet packet = in.readPacket();
				System.out.print("Captured packet #"+i+": "+packet+" ... "
				/*
				 +":"+new BigInteger(((EthernetPacket)packet).getEthernetHeader()).toString(16)
				 */
					);
				out.writePacket(packet);
				System.out.println("Dumped packet #"+i);
			}
			catch (SocketTimeoutException exc)
			{
				System.out.println(exc.getMessage());
				continue;
			}
		}
		out.flush();
		System.out.println("Flushed the dump file");
		long	bytes = out.getFilePointer();
		System.out.println("Wrote "+bytes+" bytes to the dump file");
		out.close();
		System.out.println("Closed the dump file");
		pc.close();
		System.out.println("Closed the captor");
	}
	
	// test captureTo() functions to file
	void	testCaptureTo() throws JpcapException, IOException
	{
		PacketCapture	pc = new PacketCapture(file);
		System.out.println("opened captor for dump file "+file);
		
		String	filter = "icmp";
		pc.setFilter(filter, true);
		
		File	outFile = new File("icmp_filtered.cap");
		PacketOutputStream	out = pc.getOutputStream(outFile);
		
		System.out.println("Capturing from "+file+" to "+out+" with filter "+filter);
		
		int rc = pc.captureTo(-1, out);
		
		System.out.println("Capture finished: count="+rc);
		
		out.close();
		System.out.println("Closed the filtered dump file");
		
		System.out.println("Filtered dump file should be shorter than the input dump file, as it only contains icmp.");
		
		pc.close();
		System.out.println("Closed the captor");
		
	}
	
	/**
	 * Try to read the dump file all the way to EOF
	 */
	void	testCaptureNext() throws CaptureFileOpenException, CapturePacketException
	{
		PacketCapture	pc = null;
		try
		{
			pc = new PacketCapture(file);
			System.out.println("Opened "+file+" for captureNext(), file.length="+file.length());
			for (;;)
			{
				Packet	packet = pc.captureNext();
				System.out.println("captureNext() returned packet: "+packet);
			}
		}
		catch (EOFException exc)
		{
			System.out.println("EOFException received OK");
		}
		catch (IOException exc)
		{
			System.out.println("IOException received: "+exc);
			exc.printStackTrace(System.out);
		}
		finally
		{
			if (pc != null)
				pc.close();
		}
	}
	
	/**
	 * test captureTo() functions to net.
	 * You can really only see if this works if
	 * (a) it doesn't crash, and
	 * (b) the same traffic as in the file is sniffed by a sniffer.
	 */
	void	testCaptureToNet() throws JpcapException, IOException
	{
		PacketCapture	pc = new PacketCapture(file);
		System.out.println("opened captor to read dump file "+file);
		
		CaptureDevice	device = PacketCapture.findCaptureDevice();
		PacketCapture	pc2 = new PacketCapture(device, false);
		System.out.println("opened captor to send to network device "+device);
		
		PacketOutputStream	out = pc2.getOutputStream();
		
		System.out.println("Capturing from "+file+" to "+out);
		
		pc.captureTo(-1, out);
		
		System.out.println("Capture finished: "+pc.getStatistics());
		
		out.close();
		System.out.println("Closed the dump file");
		
		pc2.close();
		System.out.println("closed the network output captor");
		pc.close();
		System.out.println("Closed the savefile input captor");
		
	}
	
	void	testInstances()
	{
		final int	count = 100000;
		for (int i = 0; i < count; i++)
		{
			PacketCapture	pc = new PacketCapture();
		}
		System.gc();
		System.out.println("Created "+count+" instances of PacketCapture");
	}
	
	/**
	 * Simple test to exercise PacketCapture.
	 */
	public static void main(String [] args) throws IOException
	{
		// Just to prove that java.net.NetworkInterface has nothing to do with CaptureDevice.
//	  Enumeration<NetworkInterface>	nwifs = NetworkInterface.getNetworkInterfaces();
//	  while (nwifs.hasMoreElements())
//		  System.out.println(nwifs.nextElement());
		
		CaptureTest	ct = new CaptureTest();
		ct.runTest();
	}
	
	
	private static final String _rcsid =
		"$Id: CaptureTest.java,v 1.3 2008/08/28 04:18:59 esmondpitt Exp $";
}
