/*
 * JWinPcap.java
 *
 * Created on 16 May 2008, 14:24
 *
 * Copyright (C) 2008, Esmond Pitt                                         *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 */

package net.sourceforge.jpcap.capture;
import java.io.File;
import java.io.IOException;
import java.net.PasswordAuthentication;
import java.net.URI;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.LongBuffer;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Set;
import net.sourceforge.jpcap.net.EthernetPacket;
import net.sourceforge.jpcap.net.Packet;
import net.sourceforge.jpcap.util.Timeval;

/**
 * The JWinPcap class encapsulates the extended functions supported by the WinPcap native library.
 * For obvious reasons it is only available under Windows and any other platforms that may support them.
 * Execution on any platform that
 * doesn't support these extensions will cause an {@link java.lang.UnsupportedOperationException}.
 *
 * @author Esmond Pitt
 * @version $Revision: 1.2 $
 *
 * @since 2.0.0
 */
public class JWinPcap extends PacketCapture
{
	/**
	 * Default buffer size for output streams that write to the network.
	 * #see #getOutputStream()
	 */
	public static final int	DEFAULT_BUFFER_SIZE = 8192;
	/**
	 * Internal representation of the type of source in use (file, remote/local interface).
	 */
	/*
	 * This indicates a file: the user want to open a capture from a local file.
	 */
	public static final int	PCAP_SRC_FILE		= 2;
	/**
	 * This indicates a local interface: the user want to open a capture from a local interface.
	 * This does not involve the RPCAP protocol.
	 */
	public static final int	PCAP_SRC_IFLOCAL	= 3;
	/**
	 * This indicates a remote interface: the user want to open a capture from an interface on a remote host.
	 * This does involve the RPCAP protocol.
	 */
	public static final int	PCAP_SRC_IFREMOTE	= 4;
	/**
	 * This string indicates that the user wants to open a capture from a local file.
	 */
	public static final String	 	PCAP_SRC_FILE_STRING	= "file://";
	/**
	 * This string indicates that the user wants to open a capture from a network interface.
	 * This string does not necessarily involve the use of the RPCAP protocol.
	 * If the interface required resides on the local host, the RPCAP protocol is not involved and the local functions are used.
	 */
	public static final String	 	 	PCAP_SRC_IF_STRING	= "rpcap://";
	
	// pcap_open flags
	/**
	 * Open flags, for use with {@link #open open()}.
	 * @see #open
	 */
	public enum OpenFlags
	{
		/**
		 * Defines if the adapter has to go in promiscuous mode.
		 */
		Promiscuous,
		/**
		 * Defines if the data transfer (in case of a remote capture) has to be done with UDP protocol.
		 */
		DataTxUDP,
		/**
		 * Defines if the remote probe will capture its own generated traffic.
		 * Note that by default, it will do so.
		 */
		NoCaptureRemote,
		/**
		 * Defines if the local adapter will capture its own generated traffic.
		 * Note that by default, it will do so.
		 */
		NoCaptureLocal,
		/**
		 * This flag configures the adapter for maximum responsiveness.
		 */
		MaxResponsiveness;
	}
	
	// Instance data
	/**
	 * The initial buffer size is 1 megabyte (see the Winpcap 4.02 documentation), then adjusted via setBufferSize().
	 * @see #getBufferSize
	 * @see #setBufferSize
	 */
	int	bufferSize = 1024*1024;
	/**
	 * The initial minToCopy value is 16000: see the Winpcap mailing list 29 May 2008.
	 * @see #getMinimumKernelCopy
	 * @see #setMinimumKernelCopy
	 */
	int minToCopy = 16000;
	/**
	 * StatisticsListeners
	 */
	private Set<StatisticsListener>	statisticsListeners = new HashSet<StatisticsListener>();
	
	/**
	 * Creates a new default instance of JWinPcap.
	 */
	public JWinPcap()
	{
		super();
	}
	
	/**
	 * Creates a new instance of JWinPcap with default values for <code>snaplen</code> and <code>timeout</code>.
	 * @param device Capture device as returned by {@link #findCaptureDevice} or {@link #lookupCaptureDevices}.
	 * @param flags Open flags, see {@link OpenFlags}
	 * @param auth Password authentication object, or null.
	 * @exception CaptureDeviceOpenException Problem opening the specified capture device.
	 */
	public JWinPcap(CaptureDevice device, EnumSet<OpenFlags>  flags, PasswordAuthentication auth)
	throws CaptureDeviceOpenException
	{
		this(device, DEFAULT_SNAPLEN, flags, DEFAULT_TIMEOUT, auth);
	}
	
	/**
	 * Creates a new instance of JWinPcap
	 * @param device Capture device as returned by {@link #findCaptureDevice} or {@link #lookupCaptureDevices}.
	 * @param snaplen Snapshot length
	 * @param flags Open flags, see {@link OpenFlags}
	 * @param timeout Timeout before packet driver starts notifying packet listeners
	 * @param auth Password authentication object, or null.
	 * @exception CaptureDeviceOpenException Problem opening the specified capture device.
	 */
	public JWinPcap(CaptureDevice device, int snaplen, EnumSet<OpenFlags> flags, int timeout, PasswordAuthentication auth)
	throws CaptureDeviceOpenException
	{
		super();
		open(device.getName(), snaplen, flags, timeout, auth);
	}
	
	/**
	 * Creates a new instance of JWinPcap
	 * @param file <i>tcpdump</i>-formatted savefile to capture from
	 */
	public JWinPcap(File file)
	throws CaptureDeviceOpenException
	{
		super();
		// NB File.toURI() returns file:/, we need file://
		open(PCAP_SRC_FILE_STRING+file.getPath(), DEFAULT_SNAPLEN, EnumSet.noneOf(OpenFlags.class), DEFAULT_TIMEOUT, null);
	}
	
	int	captureToNetwork(int count, long outHandle) throws CapturePacketException, IOException, IllegalArgumentException
	{
		return captureToNetwork0(handle, count, outHandle);
	}
	
	native int	captureToNetwork0(long handle, int count, long outHandle) throws CapturePacketException, IOException, IllegalArgumentException;
	
	@Override
		public synchronized void	close()
	{
		super.close();
	}
	
	/**
	 * Returns the first non-loopback device found by {@link #lookupCaptureDevices}.
	 *
	 * @return a CaptureDevice.
	 *
	 * TODO there should be a version of this that takes host/port/protocol.
	 */
	public static CaptureDevice	findCaptureDevice() throws CaptureDeviceNotFoundException, CaptureConfigurationException
	{
		try
		{
			for (CaptureDevice device : lookupCaptureDevices())
				if (!device.isLoopback() && device.getNetwork() != 0)
					return device;
		}
		catch (CaptureDeviceLookupException exc)
		{
			throw new CaptureDeviceNotFoundException(exc);
		}
		throw new CaptureDeviceNotFoundException("no usable capture devices detected");
	}
	
	/**
	 * {@inheritDoc}
	 *<p>
	 * This override calls {@link #getOutputStream(int, boolean)},
	 * with the values {@link #DEFAULT_BUFFER_SIZE} and <code>false</code>,
	 * returning a buffered PacketOutputStream.
	 *
	 * @see #getOutputStream(int, boolean)
	 *
	 */
	@Override
		public PacketOutputStream	getOutputStream() throws IOException
	{
		return getOutputStream(DEFAULT_BUFFER_SIZE, false);
	}
	
	/**
	 * Return a buffered packet output stream.
	 * This is a wrapper for the WinPcap <code>pcap_sendqueue_*</code> family of functions.
	 * The output stream is buffered and auto-flushed when <code>bufferSize</code> is reached.
	 *
	 * This stream is more efficient than the {@link PacketOutputStream} returned by {@link PacketCapture#getOutputStream()},
	 * because the packets are buffered in the pcap_sendqueue at the JNI/WinPcap level,
	 * before being passed to the kernel driver in a single operation
	 * when {@link PacketOutputStream#flush} is called (whether internally or by the application),
	 * so the number of context switches can be reduced considerably.
	 * Therefore, you can expect much better packet throughput.
	 *
	 * @param bufferSize Size of underlying packet queue.
	 * @param timed If true, packets are sent respecting the intervals between the packet timestamps;
	 * if false, they are sent as fast as possible.
	 * @return a buffered PacketOutputStream
	 *
	 * @since 2.0.0.
	 */
	public synchronized PacketOutputStream	getOutputStream(int bufferSize, boolean timed) throws IOException
	{
		if (!isSupported())
			throw new UnsupportedOperationException("JWinPcap class is not available on this platform");
		if (networkOutputStream == null)
			networkOutputStream = new BufferedPacketOutputStream(handle, bufferSize, timed);
		return networkOutputStream;
	}
	
	/**
	 * Return true iff the specified filter would filter the specified packet.
	 *
	 * @param filter Filter string
	 * @param packet Packet to test
	 * @return true if the filter filters the packet.
	 */
	public boolean	isFilteredOffline(CompiledFilter filter, Packet packet)
	throws InvalidFilterException
	{
		if (!isSupported())
			throw new UnsupportedOperationException("JWinPcap class is not available on this platform");
		EthernetPacket	ethernetPacket = (EthernetPacket)packet;
		byte[]	header = ethernetPacket.getEthernetHeader();
		byte[]	data = ethernetPacket.getEthernetData();
		ByteBuffer	bb = ByteBuffer.allocateDirect(header.length+data.length);
		bb.put(header);
		bb.put(data);
		return isFilteredOffline0(filter.getHandle(), bb);
	}
	
	// native support
	/**
	 * Return true if the supplied packet would be filtered by the specified compiled filter.
	 * @param filter Handle to compiled C++ filter object
	 * @param bb ByteBuffer containing the packet, may be direct or non-direct.
	 * @return true if the filter filters the packet.
	 */
	private native boolean	isFilteredOffline0(long filter, ByteBuffer bb)
	throws InvalidFilterException;
	
	/**
	 * Return true iff the other methods of this class are supported on this platform.
	 * @return true or false.
	 */
	public static boolean	isSupported()
	{
		return isSupported0();
	}
	
	// native support
	private static native boolean	isSupported0();
	
	/**
	 * Lookup capture devices.
	 * The CaptureDevice objects returned by this call can only be used with constructors and methods of this class
	 * where they override methods of {@link PacketCapture}.
	 */
	public static CaptureDevice[]	lookupCaptureDevices() throws CaptureDeviceLookupException, CaptureConfigurationException
	{
		return lookupCaptureDevices0(PCAP_SRC_IF_STRING, PCAP_SRC_IFLOCAL);
	}
	
	/**
	 * Lookup capture devices.
	 * The CaptureDevice objects returned by this call can only be used with constructors and methods of this class.
	 * @param uri URI, of the form protocol:// host, or host:port, or file://directory.
	 */
	public static CaptureDevice[]	lookupCaptureDevices(URI uri) throws CaptureDeviceLookupException, CaptureConfigurationException
	{
//		System.out.println("JWinPcap.lookupCaptureDevices(): URI="+uri);
		if (!isSupported())
			throw new UnsupportedOperationException("JWinPcap class is not available on this platform");
		int	type;
		if (uri.getScheme().equalsIgnoreCase("file"))
			type = PCAP_SRC_FILE;
		else if (uri.getHost() != null)
			type = PCAP_SRC_IFREMOTE;
		else
			type = PCAP_SRC_IFLOCAL;
		return lookupCaptureDevices0(uri.toString(), type);
	}
	
	private static native CaptureDevice[]	lookupCaptureDevices0(String uri, int type) throws CaptureDeviceLookupException, CaptureConfigurationException;
	
	@Override
		public void	open(String name, boolean promiscuous) throws CaptureDeviceOpenException
	{
		EnumSet<OpenFlags>	flags = promiscuous ? EnumSet.of(OpenFlags.Promiscuous) : EnumSet.noneOf(OpenFlags.class);
		open(name, DEFAULT_SNAPLEN, promiscuous, DEFAULT_TIMEOUT);
	}
	
	/**
	 * {@inheritDoc}
	 *<p>
	 * Calling this method is equivalent to calling
	 * <code>open(name, snaplen, flags, timeout, null)</code>
	 * where <code>flags</code> is either
	 * <code>EnumSet.of(OpenFlags.Promiscuous)</code> or <code>EnumSet.noneOf(OpenFlags.class)</code>
	 * depending on whether <code>promiscuous</code> is true or false.
	 */
	@Override
		public void	open(String name, int snaplen, boolean promiscuous, int timeout) throws CaptureDeviceOpenException
	{
		EnumSet<OpenFlags>	flags = promiscuous ? EnumSet.of(OpenFlags.Promiscuous) : EnumSet.noneOf(OpenFlags.class);
		open(name, snaplen, flags, timeout, null);
	}
	
	/**
	 * Open a capture device or file.
	 * @param name Device name as returned by CaptureDevice.getName(),
	 * or a file:// URI string of a <i>tcpdump</i>-formatted savefile.
	 * @param snaplen Snapshot length
	 * @param flags Open flags
	 * @param timeout Timeout before calling the listener, to allow the kernel to pass more
	 * than one packet into user space at a time
	 * @param auth PasswordAuthentication
	 *
	 * @exception CaptureDeviceOpenException Problem opening the specified device or file.
	 */
	public void	open(String name, int snaplen, EnumSet<OpenFlags> flags, int timeout, PasswordAuthentication auth)
	throws CaptureDeviceOpenException
	{
		if (!isSupported())
			throw new UnsupportedOperationException("JWinPcap class is not available on this platform");
		int	binaryFlags = 0;
		for (OpenFlags flag : flags)
			// NB we are relying on the ordinals mapping to the C values correctly here.
			binaryFlags |= 1 << flag.ordinal();
		String username = null;
		String	password = null;
		if (auth != null)
		{
			username = auth.getUserName();
			password = new String(auth.getPassword());
		}
		super.linkType = open0(handle, name, snaplen, binaryFlags, timeout, username, password);
		isClosed = false;
		super.snaplen = snaplen;
		try
		{
			super.netmask = getNetmask(name);
		}
		catch (CaptureConfigurationException exc)
		{
			// ignored, bad luck, no netmask, no filtering on IPv4 broadcast addresses.
			super.netmask = 0;
		}
		if (name.startsWith(PCAP_SRC_FILE_STRING))
			super.inputFile = new File(name);
	}
	
	// native support
	private native int	open0(long handle, String name, int snaplen, int flags, int timeout, String username, String password) throws CaptureDeviceOpenException;
	
	@Override
		public void	openOffline(String filename) throws CaptureFileOpenException
	{
		try
		{
			open(PCAP_SRC_FILE_STRING+filename, false);
		}
		catch (CaptureDeviceOpenException exc)
		{
			throw new CaptureFileOpenException(exc);
		}
	}
	
	/**
	 * Return the kernel buffer size.
	 * @return the kernel buffer size.
	 *
	 * @see #setBufferSize(int)
	 */
	public int	getBufferSize()
	{
		return bufferSize;
	}
	
	/**
	 * Set the size of the kernel buffer associated with an adapter.
	 * @param bufferSize specifies the size of the buffer in bytes.
	 * If an old buffer was already created with a previous call to setBufferSize(),
	 * it is deleted and its content is discarded. {@link #open} creates a 1 MByte buffer by default.
	 *
	 * @see #getBufferSize()
	 */
	public void	setBufferSize(int bufferSize) throws IOException
	{
		if (!isSupported())
			throw new UnsupportedOperationException("JWinPcap class is not available on this platform");
		setBufferSize0(handle, bufferSize);
		this.bufferSize = bufferSize;
	}
	
	/**
	 * @return 0 when the call succeeds, -1 otherwise.
	 */
	private native int	setBufferSize0(long handle, int dim) throws IOException;
	
	/**
	 * Capture mode, for use with {@link #setCaptureMode setCaptureMode()}.
	 * @see #setCaptureMode
	 */
	public enum CaptureMode
	{
		/**
		 * Packet capture mode.
		 */
		Capture,
		/**
		 * Statistics capture mode.
		 * In statistical mode, {@link StatisticsEvent StatisticsEvents} are delivered to a
		 * {@link StatisticsListener}.
		 * You can use a StatisticsListener to compute packets per second
		 * and bytes or bits per second, as shown in the tutorial
		 * linked below.
		 *
		 * @see JWinPcap#setCaptureMode
		 * @see JWinPcap#addStatisticsListener
		 * @see JWinPcap#removeStatisticsListener
		 * @see <a href="http://www.winpcap.org/docs/docs_40_2/html/group__wpcap__tut9.html">
		 * Gathering Statistics on the network traffic
		 * </a>
		 */
		Statistical,
		/**
		 * Undocumented & unknown.
		 */
		Monitor;
	}
	
	private CaptureMode	captureMode = CaptureMode.Capture;
	
	/**
	 * Return the current capture mode.
	 * @return the capture mode.
	 */
	public CaptureMode	getCaptureMode()
	{
		return captureMode;
	}
	
	/**
	 * Set the capture mode.
	 * @param mode Capture mode
	 */
	public void	setCaptureMode(CaptureMode mode)
	 throws IllegalArgumentException
	{
		if (!isSupported())
			throw new UnsupportedOperationException("JWinPcap class is not available on this platform");
		setCaptureMode0(handle, mode.ordinal());
		this.captureMode = mode;
	}
	
	// native support
	private native void	setCaptureMode0(long handle, int mode) throws IllegalArgumentException;
	
	// Listener support
	/**
	 * Add a StatisticsListener.
	 * This method is thread-safe, and safe to call during a
	 * {@link StatisticsListener#statisticsArrived StatisticsListener.statisticsArrived()} callback.
	 * @param listener Listener to add
	 */
	public synchronized void	addStatisticsListener(StatisticsListener listener)
	{
		synchronized (statisticsListeners)
		{
			if (!statisticsListeners.add(listener))
				throw new IllegalArgumentException("listener already registered");
		}
	}
	
	/**
	 * Remove a StatisticsListener.
	 * This method is thread-safe, and safe to call during a
	 * {@link StatisticsListener#statisticsArrived StatisticsListener.statisticsArrived()} callback.
	 * @param listener Listener to remove
	 */
	public synchronized void	removeStatisticsListener(StatisticsListener listener)
	{
		synchronized (statisticsListeners)
		{
			if (!statisticsListeners.remove(listener))
				throw new IllegalArgumentException("listener not registered");
		}
	}

	/**
	 * Dispatch a StatisticsEvent.
	 * @param event StatisticsEvent.
	 */
	void	dispatchStatisticsEvent(StatisticsEvent event)
	{
		Set<StatisticsListener> listeners;
		synchronized (this.statisticsListeners)
		{
			listeners = new HashSet<StatisticsListener>(statisticsListeners);
		}
		for (StatisticsListener listener : listeners)
			listener.statisticsArrived(event);
	}
	
	/**
	 * {@inheritDoc}
	 * <p>
	 * PacketHandler.handlePacket override, to generate statistics events.
	 */
	@Override
	protected void handlePacket(int length, int caplen, int seconds, int useconds, ByteBuffer bb)
	{
		switch (captureMode)
		{
			case Statistical:
				synchronized (bb)
				{
					bb.order(ByteOrder.nativeOrder());
					bb.position(0);
					long packets = bb.getLong();
					long bytes = bb.getLong();
					Timeval	tv = new Timeval(seconds, useconds);
					StatisticsEvent	event = new StatisticsEvent(this, packets, bytes, tv);
					dispatchStatisticsEvent(event);
				}
				break;
			default:
				super.handlePacket(length, caplen, seconds, useconds, bb);
		}
	}
	
	/**
	 * Fetch the minimum amount of data sent by the kernel in a single call.
	 * @return Minimum amount of packet data to accumulate before
	 * calling listeners.
	 * @see #setMinimumKernelCopy
	 */
	public int	getMinimumKernelCopy()
	{
		return minToCopy;
	}
	
	/**
	 * Set the minimum amount of data sent by the kernel in a single call.
	 * This method changes the minimum amount of data in the kernel buffer that causes a read from the application to return
	 * (unless the timeout expires).
	 * If the value of size is large, the kernel is forced to await the arrival of several packets before copying the data to the user.
	 * This guarantees a low number of system calls, i.e. low processor usage, and is a good setting for applications like packet-sniffers and protocol analyzers.
	 * Vice versa, in presence of a small value for this variable, the kernel will copy the packets as soon as the application is ready to receive them.
	 * This is useful for real time applications that need the best responsiveness from the kernel.
	 *<p>
	 * Note that the Winpcap documentation of the corresponding <i>pcap_setmintocopy()</i> function,
	 * from which the above is taken, is incorrect:
	 * it defines the minimum amount of data to copy <i>from</i> the kernel.
	 *
	 * @param minToCopy Minimum amount of packet data to accumulate before
	 * calling listeners.
	 *
	 * @exception IOException Problem changing the minimum kernel copy quantum.
	 * @see #getMinimumKernelCopy
	 */
	public void	setMinimumKernelCopy(int minToCopy) throws IOException
	{
		setMinimumKernelCopy0(handle, minToCopy);
		this.minToCopy = minToCopy;
	}
	
	// native support
	private native void	setMinimumKernelCopy0(long handle, int minimum) throws IOException;
	
	/**
	 * Sampling method for use with {@link #setSamplingMethod setSamplingMethod()}.
	 * @see #setSamplingMethod
	 */
	public enum	SamplingMethod
	{
		/**
		 * No sampling.
		 */
		None,
		/**
		 * Sample one packet after every N packets,
		 * where N is given by the value supplied to {@link #setSamplingMethod}.
		 */
		OneEveryN,
		/**
		 * Sample the first packet to arrive after every N milliseconds,
		 * where N is given by the value supplied to {@link #setSamplingMethod}.
		 */
		FirstAfterNms;
	}
	
	/**
	 * Get the current sampling method.
	 * @return the current sampling method, as per {@link #setSamplingMethod}.
	 */
	public SamplingMethod	getSamplingMethod()
	{
		return SamplingMethod.values()[getSamplingMethod0(handle)];
	}
	
	private native int	getSamplingMethod0(long handle);

	/**
	 * Get the current sampling value.
	 * @return the current sampling value, as per {@link #setSamplingMethod}.
	 */
	public int	getSamplingValue()
	{
		return getSamplingValue0(handle);
	}
	
	private native int	getSamplingValue0(long handle);
	
	/**
	 * Define a sampling method for packet capture.
	 * This takes effect when you have called {@link #setCaptureMode}
	 * with mode = {@link CaptureMode#Statistical}.
	 * <p>
	 * This function applies a sampling method to the packet capture process.
	 * The parameters will be applied as soon as the capture starts.
	 * <p>
	 * <b>Warning:</b>
	 * Sampling parameters cannot be changed when a capture is active.
	 * These parameters must be applied before starting the capture.
	 * If they are applied when the capture is in progress, the new settings are ignored.
	 * <p>
	 * Sampling works only when capturing data on Win32 or reading from a file.
	 * It has not been implemented on other platforms.
	 * Sampling works on remote machines provided that the probe (i.e. the capturing device) is a Win32 workstation.
	 * 
	 * @param method Sampling method.
	 * @param value Value for N in the sampling method.
	 */
	public void	setSamplingMethod(SamplingMethod method, int value)
	{
		if (!isSupported())
			throw new UnsupportedOperationException("JWinPcap class is not available on this platform");
		setSamplingMethod0(handle, method.ordinal(), value);
	}
	
	// native support
	private native void	setSamplingMethod0(long handle, int method, int value);
	
	/**
	 * BufferedPacketOutputStream is a buffered version of {@link PacketOutputStream}.
	 * Under the hood it uses the pcap_sendqueue_*() set of functions, via the {@link SendQueue} wrapper class.
	 *
	 * @author Esmond Pitt
	 * @since 2.0.0
	 */
	class BufferedPacketOutputStream extends NetworkOutputStream
	{
		private int	queueSize;
		private int	pending;
		private boolean	sync;
		private SendQueue	queue;
		private boolean	isClosed = false;
		
		BufferedPacketOutputStream(long handle, int queueSize, boolean sync) throws IOException
		{
			super(handle);
			this.queueSize = queueSize;
			this.sync = sync;
			this.queue = new SendQueue(handle, queueSize, sync);
		}
		
		public synchronized void	close() throws IOException
		{
			if (!isClosed)
			{
				flush();
				queue.destroy();
			}
			isClosed = true;
			networkOutputStream = null;
			queue = null;
		}
		
		public synchronized void	flush() throws IOException
		{
			if (isClosed)
				throw new IOException("stream is closed");
			if (pending > 0)
			{
				queue.transmit();
				// destroy and recreate the queue, it can't be reused.
				queue.destroy();
				queue = null;
			}
			pending = 0;
		}
		
		public synchronized SendQueue	getQueue()
		{
			// Lazy queue creation.
			if (queue == null)
				this.queue = new SendQueue(handle, queueSize, sync);
			return queue;
		}
		
		public synchronized void	writePacket(Packet packet) throws IOException
		{
			if (isClosed)
				throw new IOException("stream is closed");
			packet.setChecksums();
			EthernetPacket	ethernetPacket = (EthernetPacket)packet;
			byte[]	header = ethernetPacket.getEthernetHeader();
			byte[]	data = ethernetPacket.getEthernetData();
			ByteBuffer	bb = ByteBuffer.allocateDirect(header.length+data.length);
			bb.put(header);
			bb.put(data);
			Timeval	tv = packet.getTimeval();
			int	sec = 0, usec = 0;
			if (tv != null)
			{
				sec = (int)tv.getSeconds();
				usec = (int)tv.getMicroSeconds();
			}
			// Try to write
			try
			{
				getQueue().queue(bb, sec, usec);
			}
			catch (IOException exc)
			{
				flush();
				// Try once more. This time an exception will be thrown and not caught if it fails.
				getQueue().queue(bb, sec, usec);
			}
			pending += bb.position();
		}
		
	}
	
	/**
	 * SendQueue class, for efficient sending of large number of packets.
	 *
	 * This encapsures the four <code>pcap_sendqueue_*</code> functions.
	 */
	class SendQueue
	{
		// Pointer to native queue
		private long	ptr;
		private boolean	timed;
		private long	out;

		/**
		 * @param sync If true, packets are sent respecting the intervals between the packet timestamps;
		 * if false, they are sent as fast as possible.
		 */
		SendQueue(long handle, int size, boolean sync)
		{
			this.ptr = sendQueueCreate0(handle, size, sync);
		}
		
		protected void	finalize() throws Throwable
		{
			try
			{
				destroy();
			}
			finally
			{
				super.finalize();
			}
		}
		
		/**
		 * {@inheritDoc}
		 *<p>
		 * <b>Note:</b> destroying the queue does <i>not</i>transmit any untransmitted packets.
		 */
		public void	destroy()
		{
			if (ptr != 0)
				sendQueueDestroy0(ptr);
			ptr = 0;
		}
		
		/**
		 * Queue a packet to this queue.
		 * Calls {@link Packet#setChecksums} before doing so.
		 *
		 * @param packet Packet to queue
		 * @exception IOException error adding packet, e.g. queue is full.
		 */
		public int	queue(ByteBuffer bb, int sec, int usec) throws IOException
		{
			if (ptr == 0)
				throw new IOException("queue has been closed");
			return sendQueueEnqueue0(ptr, bb, sec, usec);
		}
		
		/**
		 * Transmit the queue.
		 *
		 * It's not clear yet whether transmitting the queue empties it ready for further adds.
		 *
		 * @param sync If true, packets are sent respecting the intervals between the packet timestamps;
		 * if false, they are sent as fast as possible.
		 *
		 * @return the number of bytes transmitted.
		 */
		public int	transmit() throws IOException
		{
			if (ptr == 0)
				throw new IOException("queue has been closed");
			if (isClosed())
				throw new IOException("captor device has been closed");
			return sendQueueTransmit0(handle, ptr);
		}
	}

	// Native support
	// For reasons I do not understand, these methods aren't found at runtime if declared in the inner class.
	/**
	 * Create a send queue.
	 * @param handle Handle to C++ JWinPcap object
	 * @param size Queue size in bytes
	 * @param sync Whether queue is syncrhonized (timed when sent according to timestamps) or not
	 * @return Handle to C++ queue object.
	 */
	private native long	sendQueueCreate0(long handle, int size, boolean sync);
	/**
	 * Write to the queue.
	 * @param handle Handle to C++ queue object
	 * @param bb ByteBuffer, may be direct or non-direct.
	 * @param seconds Seconds part of timestamp
	 * @param usecs Microseconds part of timestamp.
	 */
	private native int	sendQueueEnqueue0(long handle, ByteBuffer bb, int seconds, int usecs) throws IOException;
	/**
	 * Destroy the queue.
	 * @param ptr Handle to C++ queue object
	 */
	private native void	sendQueueDestroy0(long ptr);
	/**
	 * Transmit the queue.
	 * @param handle Handle to C++ queue object
	 * @param ptr Handle to C++ queue object
	 */
	private native int	sendQueueTransmit0(long handle, long ptr) throws IOException;
}
