// $Id: JWinPcapTest.java,v 1.1 2008/08/18 05:34:36 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Copyright (C) 2008, Esmond Pitt                                         *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.capture;

import java.io.File;
import java.io.IOException;
import java.net.SocketTimeoutException;
import java.net.URI;
import java.util.Arrays;
import java.util.EnumSet;
import net.sourceforge.jpcap.net.Packet;
import net.sourceforge.jpcap.net.RawPacket;

/**
 * Class to exercise JWinPcap.
 * <p>
 * Testing of JWinPcap could easily be implemented in the same class's
 * main method, but since the tester is a client and must implement the 
 * listener interface, seems clearer to decouple the two.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.1 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:36 $
 */
public class JWinPcapTest extends CaptureTest implements PacketListener, RawPacketListener, StatisticsListener
{
  protected static String NAME = "JWinPcapTest";
	private File	file = new File("test_captures.cap");

  /**
   * Create a new tester.
   */
  public JWinPcapTest() {
  }


  // implementation of PacketListener interface
  public void	packetArrived(Packet packet)
  {
    System.err.println(NAME + ": " + packet + " arrived");
  }
  
  // implementation of RawPacketListener interface

  public void rawPacketArrived(RawPacket rawPacket) {
    System.err.println(NAME + ": " + rawPacket + " arrived");
  }
  
  public void	statisticsArrived(StatisticsEvent event)
  {
	  System.err.println(NAME+": "+event);
  }



  // implementation of test logic

  /**
   * Execute the test.
   */
  public void runTest() {
  
	  System.out.println("Library version: "+JWinPcap.getLibVersion());
	
	// Test offline functions
	try
	{
		testLookupDevice();
		testDump();
		testCaptureTo();
		testCaptureToNet();
	}
	catch (Exception exc)
	{
		exc.printStackTrace();
		System.exit(1);
	}
	
  }

  // Exercise same as above but via the findLookupDevice() method,
  // which also exercises listLookupDevices().
  void	testLookupDevice() throws Exception
  {
	  // test lookup capture files
	  try
	  {
		  CaptureDevice[]	cfs = JWinPcap.lookupCaptureDevices(new URI("file://c:/java/jpcap-0.01.16/"));
		  System.out.println("Found "+cfs.length+" capture files");
		  for (CaptureDevice cf : cfs)
			  System.out.println("Found capture file "+cf.getName());
	  }
	  catch (JpcapException exc)
	  {
		  exc.printStackTrace();
	  }
	  
    JWinPcap pc = null;

    //-- set the capture device
    CaptureDevice device = null;
    try {
      System.err.print(NAME + ": detecting capture device.. ");
      device = JWinPcap.findCaptureDevice();
      System.err.println(device+": network="+Integer.toHexString(device.getNetwork())+":netmask="+Integer.toHexString(device.getNetmask()));
    }
    catch(CaptureDeviceNotFoundException e) {
      // usually occurs if no network devices are available on the system
      System.err.println(e);
      System.exit(1);
    }
    catch(CaptureConfigurationException e) {
      // Shouldn't happen unless the device name is wrong
      System.err.println(e);
      System.exit(1);
    }

    //-- open the capture device
    try {
      System.err.print(NAME + ": opening capture device.. ");
      pc = new JWinPcap(device, EnumSet.of(JWinPcap.OpenFlags.Promiscuous), null);
      System.err.println(device);
    }
    catch(CaptureDeviceOpenException e) {
      // usually occurs if the user doesn't have sufficient privileges
      System.err.println(e);
      System.exit(1);
    }
	
	try
	{
		// Exercise link layer type API
		System.err.println("Link layer name="+pc.getLinkLayerName());
		System.err.println("Link layer description="+pc.getLinkLayerDescription());
		
		// Exercise sampling API
		System.err.println("Capture mode="+pc.getCaptureMode());
		System.err.println("Sampling method="+pc.getSamplingMethod());
		System.err.println("Sampling value="+pc.getSamplingValue());
		
		pc.setCaptureMode(JWinPcap.CaptureMode.Statistical);
		pc.addStatisticsListener(this);
		pc.setSamplingMethod(JWinPcap.SamplingMethod.OneEveryN, 1);

		System.err.println("Capture mode="+pc.getCaptureMode());
		System.err.println("Sampling method="+pc.getSamplingMethod());
		System.err.println("Sampling value="+pc.getSamplingValue());
	}
	catch (CaptureConfigurationException exc)
	{
      exc.printStackTrace();
      System.exit(1);
	}

    //-- set filters
    try {
      String filter = ""; // nothing filtered by default
      System.err.println(NAME + ": setting filter to '" + filter + "'.. ");
      pc.setFilter(filter, true);
      System.err.println(NAME + ": filter compiled and activated ok");
    }
    catch(InvalidFilterException e) {
      System.err.println(e);
      System.exit(1);
    }

	// Set the kernel buffer and min to copy
	System.err.print(NAME+": setting buffer size to 512k ...");
	pc.setBufferSize(512*1024);
	System.err.println("ok");
	System.err.println("buffer size now="+pc.getBufferSize());
	System.err.print(NAME+": setting minToCopy to 256 bytes ...");
	pc.setMinimumKernelCopy(512);
	System.err.println("ok");

    //-- register the tester as a packet listener
    System.err.print(NAME + ": registering as a packet listener.. ");
    pc.addPacketListener(this);
    System.err.println("ok");

	//-- register the tester as a raw packet listener
    System.err.print(NAME + ": registering as a raw packet listener.. ");
    pc.addRawPacketListener(this);
    System.err.println("ok");

    //-- capture packets
    try {
      int count = TEST_COUNT;
      System.err.println(NAME + ": waiting for " + count + " packet(s).. ");
      pc.capture(count);
      System.err.println(NAME + ": done capturing.");
    }
    catch(CapturePacketException e) {
		e.printStackTrace();
      System.exit(1);
    }

    //-- dump statistics
    System.err.println(pc.getStatistics());	  
	
	pc.close();
  }
  
// Exercise offline functions
  public void	testDump() throws Exception
  {
	CaptureDevice[]	devices = JWinPcap.lookupCaptureDevices();
	System.out.println("Devices="+Arrays.asList(devices));
	// Open the first device in the list.
	// This API doesn't return 'down' devices, so that should be OK.
	JWinPcap	pc = new JWinPcap(devices[0], EnumSet.of(JWinPcap.OpenFlags.Promiscuous), null);
	System.out.println("opened "+devices[0]);
	
	// Test output stream
	PacketOutputStream	out = pc.getOutputStream(file);
	System.out.println("Opened dump file "+file);
	
	// Test input stream
	PacketInputStream	in = pc.getInputStream();
	System.out.println("got packet input stream");
	for (int i = 0; i < TEST_COUNT; i++)
	{
		try
		{
			Packet packet = in.readPacket();
			System.out.println("captured packet #"+(i+1)+": "+packet);
			out.writePacket(packet);
			System.out.println("Dumped packet #"+(i+1));
		}
		catch (SocketTimeoutException exc)
		{
			System.err.println(exc.getMessage());
			break;
		}
	}
	out.flush();
	System.out.println("Flushed the dump file");
	long	bytes = out.getFilePointer();
	System.out.println("Wrote "+bytes+" bytes to the dump file");
	out.close();
	out.close();
	System.out.println("Closed the dump file");
	pc.close();
	System.out.println("Closed the captor");
  }
  
  // test captureTo() functions to file
  void	testCaptureTo() throws JpcapException, IOException
  {
	JWinPcap	pc = new JWinPcap(file);
	System.out.println("opened captor for dump file "+file);
	
	File	outFile = new File("captureTo_out.cap");
	PacketOutputStream	out = pc.getOutputStream(outFile);
	
	System.out.println("Capturing from "+file+" to "+out);

	pc.captureTo(-1, out);
	
	System.out.println("Capture finished: "+pc.getStatistics());
	
	out.close();
	System.out.println("Closed the dump file");
	
	pc.close();	
	System.out.println("Closed the captor");
	  
  }
  
  /**
   * test captureTo() functions to net.
   * You can really only see if this works if
   * (a) it doesn't crash, and
   * (b) the same traffic as in the file is sniffed by a sniffer.
   */
  void	testCaptureToNet() throws JpcapException, IOException
  {
	JWinPcap	pc = new JWinPcap(file);
	System.out.println("opened captor to read dump file "+file);
	
	CaptureDevice	device = JWinPcap.findCaptureDevice();
	JWinPcap	pc2 = new JWinPcap(device, EnumSet.of(JWinPcap.OpenFlags.Promiscuous), null);
	System.out.println("opened captor to send to network device "+device);
	
	PacketOutputStream	out = pc2.getOutputStream();
	
	System.out.println("Capturing from "+file+" to "+out);

	int	count = pc.captureTo(-1, out);
	
	System.out.println("Capture finished count="+count+": "+pc.getStatistics());
	
	out.close();
	System.out.println("Closed the dump file");

	pc2.close();
	System.out.println("closed the network output captor");
	pc.close();	
	System.out.println("Closed the savefile input captor");
	  
  }


  /**
   * Simple test to exercise JWinPcap.
   */
  public static void main(String [] args) throws IOException {
	  // Just to prove that java.net.NetworkInterface has nothing to do with CaptureDevice.
//	  Enumeration<NetworkInterface>	nwifs = NetworkInterface.getNetworkInterfaces();
//	  while (nwifs.hasMoreElements())
//		  System.out.println(nwifs.nextElement());
	  
    JWinPcapTest ct = new JWinPcapTest();
    ct.runTest();
  }


  private static final String _rcsid = 
    "$Id: JWinPcapTest.java,v 1.1 2008/08/18 05:34:36 esmondpitt Exp $";
}
