// $Id: ICMPPacket.java,v 1.2 2008/08/18 05:34:37 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.net;

import net.sourceforge.jpcap.util.AnsiEscapeSequences;
import net.sourceforge.jpcap.util.Timeval;
import java.nio.ByteBuffer;


/**
 * An ICMP packet.
 * <p>
 * Extends an IP packet, adding an ICMP header and ICMP data payload.
 *
 * @author Patrick Charles
 * @author Jonas Lehmann
 * @author Esmond Pitt
 *
 * @version $Revision: 1.2 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:37 $
 */
public class ICMPPacket extends IPPacket implements ICMPFields, IP.ICMP
{
	static final long serialVersionUID = -4164595699639702922L;
	
 /**
   * Construct a new ICMP packet.
   *
   * @param lLen link-layer header length
   * @param bytes Packet data
   */
  public ICMPPacket(int lLen, byte [] bytes) {
    this(lLen, bytes, null);
  }

  /**
   * Construct a new ICMP packet, including the capture time.
   *
   * @param lLen link-layer header length
   * @param bytes Packet data
   * @param tv Timestamp
   */
  public ICMPPacket(int lLen, byte [] bytes, Timeval tv) {
    super(lLen, bytes, tv);
  }

  /**
   * Construct a new ICMP packet, including the capture time.
   *
   * @param lLen link-layer header length
   * @param bb Packet data in a ByteBuffer
   * @param tv Timestamp
   * @since 2.0.0
   */
  public ICMPPacket(int lLen, ByteBuffer bb, Timeval tv) {
    super(lLen, bb, tv);
  }

  public byte[] getICMPHeader() {
	byte[]	bytes = new byte[ICMP_HEADER_LEN];
	_byteBuffer.position(_ipOffset);
	_byteBuffer.get(bytes);
	return bytes;
  }

  /**
   * {@inheritDoc}
   */
  public byte[] getHeader() {
    return getICMPHeader();
  }

  public byte[] getICMPData() {
	_byteBuffer.position(_ipOffset+ICMP_HEADER_LEN);
 	byte[]	bytes = new byte[_byteBuffer.remaining()];
	_byteBuffer.get(bytes);
	return bytes;
  }

  /**
   * {@inheritDoc}
   */
  public byte[] getData() {
    return getICMPData();
  }

  public int getMessageCode() {
	return _byteBuffer.getShort(_ipOffset+ICMP_CODE_POS) & 0xffff;
  }

  public void	setMessageCode(short code)
  {
	  _byteBuffer.putShort(_ipOffset+ICMP_CODE_POS, code);
  }
  
  /**
   * {@inheritDoc}
   * <p>
   * Formerly {@link #getMessageType()}.
   */
  public int getMessageMajorCode() {
    return _byteBuffer.get(_ipOffset+ICMP_CODE_POS) & 0xff;
  }

  /**
   * For backward compatibility.
   * @deprecated use {@link #getMessageMajorCode()}.
   */
  @Deprecated
  public int getMessageType() {
    return getMessageMajorCode();
  }

  public int getMessageMinorCode() {
    return _byteBuffer.get(_ipOffset+ICMP_CODE_POS+1) & 0xff;
  }

  /**
   * {@inheritDoc}
   */
  public int getChecksum() {
    return getICMPChecksum();
  }

  public int getICMPChecksum() {
    return _byteBuffer.getShort(_ipOffset+ICMP_CSUM_POS) & 0xffff;
  }
  
  /**
   * {@inheritDoc}
   */
  public boolean	isValidChecksum()
  {
	  return super.isValidChecksum()
	  && getICMPChecksum() == computeSenderChecksum(_ipOffset, ICMP_HEADER_LEN+getICMPData().length, ICMP_CSUM_POS);
  }
  
  /**
   * {@inheritDoc}
   */
  public void	setChecksums()
  {
	  int	checksum = computeSenderChecksum(_ipOffset, ICMP_HEADER_LEN+getICMPData().length, ICMP_CSUM_POS);
	  // Forget cache. NB getICMPHeader() re-establishes it so can't be called after this.
	  _byteBuffer.putShort(_ipOffset+ICMP_CSUM_POS, (short)checksum);
	  super.setChecksums();
  }

  /**
   * {@inheritDoc}
   */
  public String toColoredString(boolean colored) {
    StringBuffer buffer = new StringBuffer();
    buffer.append('[');
    if(colored) buffer.append(getColor());
    buffer.append("ICMPPacket");
    if(colored) buffer.append(AnsiEscapeSequences.RESET);
    buffer.append(": ");
    buffer.append(ICMPMessage.getDescription(getMessageCode()));
    buffer.append(", ");
    buffer.append(getSourceInetAddress().getHostAddress() + "->" + 
                  getDestinationInetAddress().getHostAddress());
    buffer.append(" l=" + ICMP_HEADER_LEN + "," + 
		  (_byteBuffer.limit()-_ipOffset-ICMP_HEADER_LEN));
    buffer.append(']');

    return buffer.toString();
  }

  /**
   * {@inheritDoc}
   */
  public String getColor() {
    return AnsiEscapeSequences.LIGHT_BLUE;
  }

  private static final String _rcsid = 
    "$Id: ICMPPacket.java,v 1.2 2008/08/18 05:34:37 esmondpitt Exp $";
}
