/*
 * IPv6.java
 *
 * Created on 28 May 2008, 17:22
 *
 * Copyright (c) Esmond Pitt, 2007. All rights reserved.
 */

package net.sourceforge.jpcap.net;

import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.nio.BufferUnderflowException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import net.sourceforge.jpcap.util.AnsiEscapeSequences;
import net.sourceforge.jpcap.util.Timeval;

/**
 * IPv6 interfaces and packet classes.
 *
 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc2460.txt">RFC 2460</a>
 * @since 2.0.0
 *
 * @author Esmond Pitt
 * @version $Revision: 1.1 $
 */
/*public*/interface IPv6 extends IPProtocols
{

	/**
	 * Width in bytes of an IPv6 address.
	 *
	 * @since 2.0.0
	 */
	public static final int	ADDRESS_WIDTH = 16;

	/**
	 * IPv6 field offsets.
	 * @since 2.0.0
	 *
	 */
	/*
3.  IPv6 Header Format

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |Version| Traffic Class |           Flow Label                  |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |         Payload Length        |  Next Header  |   Hop Limit   |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                                                               |
   +                                                               +
   |                                                               |
   +                         Source Address                        +
   |                                                               |
   +                                                               +
   |                                                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                                                               |
   +                                                               +
   |                                                               |
   +                      Destination Address                      +
   |                                                               |
   +                                                               +
   |                                                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   Version              4-bit Internet Protocol version number = 6.
   Traffic Class        8-bit traffic class field.  See section 7.

   Flow Label           20-bit flow label.  See section 6.

   Payload Length       16-bit unsigned integer.  Length of the IPv6
                        payload, i.e., the rest of the packet following
                        this IPv6 header, in octets.  (Note that any
                     extension headers [section 4] present are
                        considered part of the payload, i.e., included
                        in the length count.)

   Next Header          8-bit selector.  Identifies the type of header
                        immediately following the IPv6 header.  Uses the
                        same values as the IPv4 Protocol field [RFC-1700
                        et seq.].

   Hop Limit            8-bit unsigned integer.  Decremented by 1 by
                        each node that forwards the packet. The packet
                        is discarded if Hop Limit is decremented to
                        zero.

   Source Address       128-bit address of the originator of the packet.
                        See [ADDRARCH].

   Destination Address  128-bit address of the intended recipient of the
                        packet (possibly not the ultimate recipient, if
                        a Routing header is present).  See [ADDRARCH]
                        and section 4.4.

	 */
	int	IPv6_VERSION_POS = 0,	// 4 bits
		IPv6_TC_POS = 0,		// next 8 bits
		IPv6_FLOWLABEL_POS = 0,	// bottom 20 bits of first 32-bit word
		IPv6_PAYLOAD_LENGTH_POS = 4,
		IPv6_NEXTHEADER_POS = 6,
		IPv6_HOPLIMIT_POS = 7,
		IPv6_SRC_POS = 8,
		IPv6_DST_POS = 24,
		IPv6_OPTION_HEADERS_POS = 40;

	int	IPv6_VERSION = 0x06;
	
	// Pseudo-header
	int	IPv6_PSEUDO_PACKET_LENGTH_POS	= 0;	// offset from start of final header
	int	IPv6_PSEUDO_NEXT_HEADER_POS		= 7;	// offset from start of final header
	
	/**
	 * IPv6 Protocols. These are the same as for IPv4: see RFC 2460 #3.
	 *
	 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc2460.txt">RFC 2460</a>
	 * @see <a href="http://www.rfc-editor.org/cgi-bin/rfcdoctype.pl?loc=RFC&letsgo=4443&type=ftp&file_format=txt">RFC 4443</a>
	 * @see <a href=ftp://ftp.rfc-editor.org/in-notes/rfc4861.txt">RFC 4861</a>
	 *
	 * @since 2.0.0
	 */
		byte HOP_BY_HOP_OPTIONS_HEADER				= 0,
			ROUTING_HEADER							= 43,
			FRAGMENT_HEADER							= 44,
			ENCAPSULATING_SECURITY_PAYLOAD_HEADER	= 50,
			AUTHENTICATION_HEADER					= 51,
			NO_NEXT_HEADER							= 59,
			DESTINATION_OPTIONS_HEADER				= 60;

		/**
		 * IPv6 Headers namespace.
		 *
		 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc2460.txt">RFC 2460</a>
		 *
		 * @since 2.0.0
		 *
		 * @author Esmond Pitt
		 * @version $Revision: 1.1 $
		 */
		interface Headers
		{
		
			/**
			 * IPv6 Header base class.
			 *
			 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc2460.txt">RFC 2460</a>
			 * @since 2.0.0
			 *
			 * @author Esmond Pitt
			 * @version $Revision: 1.1 $
			 */
			class Header
			{
				byte	type;
				byte	nextHeader;
				byte	length;
				
				Header(byte type, ByteBuffer	bb)
				{
					this.type = type;
					this.nextHeader = bb.get();
					this.length = bb.get();
				}
				
				/**
				 * Return the option type.
				 * @return the option type.
				 */
				public int	getType()
				{
					return type;
				}
				
				/**
				 * Return the option length.
				 * @return the option length.
				 */
				public int	getLength()
				{
					return length;
				}
				
				byte	getNextHeader()
				{
					return nextHeader;
				}
			}
			
			/**
			 * IPv6 HopByHopOptions.
			 *
			 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc2460.txt">RFC 2460</a>
			 * @since 2.0.0
			 *
			 * @author Esmond Pitt
			 * @version $Revision: 1.1 $
			 */
			class HopByHopOptions extends Header
			{
				byte[]	options;
				
				HopByHopOptions(ByteBuffer bb)
				{
					super(HOP_BY_HOP_OPTIONS_HEADER, bb);
					this.options = new byte[length-1];
					bb.get(options);
				}
				
				public String	toString()
				{
					return "HopByHopOptions";
				}
			}
			
			/**
			 * IPv6 Routing.
			 *
			 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc2460.txt">RFC 2460</a>
			 * @since 2.0.0
			 *
			 * @author Esmond Pitt
			 * @version $Revision: 1.1 $
			 */
			class Routing extends Header
			{
				byte	routingType;
				byte	segmentsLeft;
				byte[]	options;
				
				Routing(ByteBuffer bb)
				{
					super(ROUTING_HEADER,  bb);
					this.routingType = bb.get();
					this.segmentsLeft = bb.get();
					this.options = new byte[length-3];
					bb.get(options);
				}
				
				public String	toString()
				{
					return "Routing";
				}
			}
			
			/**
			 * IPv6 Fragment.
			 *
			 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc2460.txt">RFC 2460</a>
			 * @since 2.0.0
			 *
			 * @author Esmond Pitt
			 * @version $Revision: 1.1 $
			 */
			class Fragment extends Header
			{
				int	fragmentOffset;
				boolean	more;
				int	identification;
				
				Fragment(ByteBuffer bb)
				{
					super(FRAGMENT_HEADER,  bb);
					this.fragmentOffset = bb.getShort() & 0xffff;
					this.more = (fragmentOffset & 1) != 0;
					this.fragmentOffset >>= 3;
					this.identification = bb.getInt();
				}
				
				public String	toString()
				{
					return "Fragment";
				}
			}
			
			/**
			 * IPv6 DestinationOptions.
			 *
			 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc2460.txt">RFC 2460</a>
			 * @since 2.0.0
			 *
			 * @author Esmond Pitt
			 * @version $Revision: 1.1 $
			 */
			class DestinationOptions extends Header
			{
				byte[]	options;
				
				DestinationOptions(ByteBuffer bb)
				{
					super(DESTINATION_OPTIONS_HEADER,  bb);
					this.options = new byte[length-1];
					bb.get(options);
				}
				
				public String	toString()
				{
					return "DestinationOptions";
				}
			}
		}
	
	/**
	 * IPv6 packet
	 *
	 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc2460.txt">RFC 2460</a>
	 * @since 2.0.0
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 */
	class Packet extends IP.Packet
	{
		static final long serialVersionUID = 5637216501741768403L;

		List<Headers.Header>	headers = new LinkedList<Headers.Header>();
		int	_ipHeaderLength;
		
		/**
		 * Creates a new instance of IPv6.Packet
		 */
		public Packet(int lLen, byte[] bytes, Timeval tv)
		{
			this(lLen, ByteBuffer.wrap(bytes), tv);
		}
		
		/**
		 * Creates a new instance of IPv6.Packet
		 */
		public Packet(int lLen, ByteBuffer bb, Timeval tv)
		{
			super(lLen, bb, tv);
			int nextHeader = HeaderFactory.parseHeaders(lLen, bb, headers);
			this._ipHeaderLength = bb.position()-_ethOffset;
			assert getVersion() == IPv6_VERSION : "Invalid IPv6 version: "+getVersion();
		}
		
		public int	getIPHeaderLength()
		{
			return _ipHeaderLength;
		}
		
		/**
		 * Fetch the next-header value <i>from the IPv6 header,</i> before the extension headers.
		 * @return the next-header value.
		 */ 
		public int	getNextHeader()
		{
			return _byteBuffer.get(IPv6_NEXTHEADER_POS) & 0xff;
		}
		
		public int	getVersion()
		{
			return _byteBuffer.get(_ipOffset+IPv6_VERSION_POS) >>> 4;
		}
		
		public int	getIPv6TrafficClass()
		{
			return (_byteBuffer.getShort(_ipOffset+IPv6_TC_POS) >>> 4) & 0xffff;
		}
		
		public int	getIPv6FlowLabel()
		{
			// Bottom 20 bits
			return _byteBuffer.getInt(_ipOffset+IPv6_FLOWLABEL_POS) & 0xfffff;
		}
		
		public InetAddress getSourceInetAddress()
		{
			try
			{
				byte[]	address = new byte[ADDRESS_WIDTH];
				_byteBuffer.position(_ipOffset+IPv6_SRC_POS);
				_byteBuffer.get(address, 0, address.length);
//				System.arraycopy(_bytes,_ipOffset+IPv6_SRC_POS,address,0,address.length);
				return InetAddress.getByAddress(address);
			}
			catch (UnknownHostException exc)
			{
				// Unexpected
				exc.printStackTrace();
				return null;
			}
		}
		
		public void	setSourceInetAddress(InetAddress address)
		{
			if (!(address instanceof Inet6Address))
				throw new IllegalArgumentException("cannot assign IPv4 address to IPv6 packet");
			_byteBuffer.position(_ipOffset+IPv6_SRC_POS);
			_byteBuffer.put(address.getAddress());
		}
		
		public InetAddress getDestinationInetAddress()
		{
			try
			{
				byte[]	address = new byte[ADDRESS_WIDTH];
				_byteBuffer.position(_ipOffset+IPv6_DST_POS);
				_byteBuffer.get(address, 0, address.length);
//				System.arraycopy(_bytes,_ipOffset+IPv6_DST_POS,address,0,address.length);
				return InetAddress.getByAddress(address);
			}
			catch (UnknownHostException exc)
			{
				// Unexpected
				exc.printStackTrace();
				return null;
			}
		}
		
		public void	setDestinationInetAddress(InetAddress address)
		{
			if (!(address instanceof Inet6Address))
				throw new IllegalArgumentException("cannot assign IPv4 address to IPv6 packet");
			_byteBuffer.position(_ipOffset+IPv6_DST_POS);
			_byteBuffer.put(address.getAddress());
		}
		
		/**
		 * Return the next-header value.
		 * @return the next-header value.
		 */
		public byte	getIPv6NextHeader()
		{
			return _byteBuffer.get(_ipOffset+IPv6.IPv6_NEXTHEADER_POS);
		}
		
		/**
		 * Return the option 
		 * @return the option 
		 */
		public List<Headers.Header>	getIPv6Headers()
		{
			return headers;
		}
		
		protected int computeSenderChecksum(int start, int len, int csumPos)
		{
			// quick bad-data check
			if( csumPos >= len)
				return 0; // bad data, header too short
			// copy bytes, zero out checksum
			byte[] sbytes = new byte[len];
			_byteBuffer.position(start);
			_byteBuffer.get(sbytes, 0, len);
//			System.arraycopy(_bytes, start, sbytes, 0, len);
			// zero out any current checksum
			sbytes[csumPos] = sbytes[csumPos+1] = 0;
			// checksum should come out to zero if checksum is correct
			return ipChecksum(sbytes,0,len);
		}

		public int	getIPProtocol()
		{
			if (headers.size() > 0)
				return headers.get(headers.size()-1).getNextHeader() & 0xff;
			return getNextHeader();
		}
		
		public byte[]	getIPHeader()
		{
			return null;
		}
		
		public byte[]	getIPData()
		{
			return null;
		}
		
		public int getLength()
		{
			return _byteBuffer.limit()-_ipOffset;
		}
		
		public String toColoredVerboseString(boolean colored)
		{
			StringBuffer buffer = new StringBuffer();
			buffer.append('[');
			if(colored) buffer.append(getColor());
			buffer.append("IPv"+getVersion()+"Packet");
			if(colored) buffer.append(AnsiEscapeSequences.RESET);
			buffer.append(": ");
			buffer.append("version=" + getVersion() + ", ");
			buffer.append("hlen=" + getHeaderLength() + ", ");
			buffer.append("tos=" + getIPv6TrafficClass() + ", ");
			buffer.append("length=" + getLength() + ", ");
//			buffer.append("id=" + getId() + ", ");
//			buffer.append("flags=0x" + Integer.toHexString(getFragmentFlags()) + ", ");
//			buffer.append("offset=" + getFragmentOffset() + ", ");
//			buffer.append("ttl=" + getTimeToLive() + ", ");
			buffer.append("proto=" + getProtocol() + ", ");
			buffer.append("sum=0x" + Integer.toHexString(getChecksum()) + ", ");
			buffer.append("src=" + getSourceInetAddress().getHostAddress() + ", ");
			buffer.append("dest=" + getDestinationInetAddress().getHostAddress());
			buffer.append(']');
			
			return buffer.toString();
		}
	}

	/**
	 * IPv6 HeaderFactory class.
	 */
	class HeaderFactory
	{
		private HeaderFactory(){}
		
		/**
		 * Parse the provided ByteBuffer from the specified initial offset,
		 * populating the provided list of headers, and return the value of the final
		 * nextHeader field.
		 * @param lLen Link-layer length: offset of the start of the IPv6 payload.
		 * @param bb Packet data wrapped in a ByteBuffer
		 * @param headers Empty list of Headers
		 * @return the last nextHeader field.
		 */
		public static int	parseHeaders(int lLen, ByteBuffer bb, List<Headers.Header> headers)
		{
			byte nextHeader = bb.get(lLen+IPv6.IPv6_NEXTHEADER_POS);
			bb.position(lLen+IPv6.IPv6_OPTION_HEADERS_POS);
			for (;;)
			{
				Headers.Header header;
				switch (nextHeader)
				{
					case HOP_BY_HOP_OPTIONS_HEADER:
						header = new Headers.HopByHopOptions(bb);
						headers.add(header);
						nextHeader = header.getNextHeader();
						continue;
					case ROUTING_HEADER:
						header = new Headers.Routing(bb);
						headers.add(header);
						nextHeader = header.getNextHeader();
						continue;
					case FRAGMENT_HEADER:
						header = new Headers.Fragment(bb);
						headers.add(header);
						nextHeader = header.getNextHeader();
						continue;
					/*
					case E_SECURITY_PAYLOAD_HEADER:
						header = new Headers.SecurityPayload(bb);
						headers.add(header);
						nextHeader = header.getNextHeader();
						continue;
					case AUTHENTICATION_HEADER:
						header = new Headers.Authentication(bb);
						headers.add(header);
						nextHeader = header.getNextHeader();
						continue;
					*/
					case DESTINATION_OPTIONS_HEADER:
						header = new Headers.DestinationOptions(bb);
						headers.add(header);
						nextHeader = header.getNextHeader();
						continue;
					case NO_NEXT_HEADER:
					case ICMPV6:
//					case IGMP:
					case TCP:
					case UDP:
					default:
						return nextHeader;
				}
			}
		}
	}
	
	/**
	 * IPv6 PacketFactory class.
	 * @since 2.0.0
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 */
	class PacketFactory
	{
		// Private ctor, no instantiation please.
		private PacketFactory(){}
		
		/**
		 * Create an IPv6 packet.
		 * @param lLen Link-layer header length
		 * @param bb ByteBuffer containing packet data
		 * @param tv Timeval of packet
		 * @return an IPv6 Packet of the appropriate class: ICMP, MLD, TCP, UDP, or IPv6.Packet by default.
		 */
		public static Packet	dataToPacket(int lLen, ByteBuffer bb, Timeval tv)
		{
			List<Headers.Header>	headers = new LinkedList<Headers.Header>();
			int nextHeader = HeaderFactory.parseHeaders(lLen, bb, headers);
			switch (nextHeader)
			{
				case ICMPV6:
					return IPv6.ICMPv6.PacketFactory.dataToPacket(lLen, bb, tv);
				case TCP:
					return new IPv6.TCP.Packet(lLen, bb, tv);
				case UDP:
					return new IPv6.UDP.Packet(lLen, bb, tv);
				default:
					return new IPv6.Packet(lLen, bb, tv);
			}
		}
	}
	
	/**
	 * ICMPv6 protocol. Note that this not just ICMP over IPv6, it is an extension of the ICMP protocol.
	 *
	 * @see <a href="http://www.rfc-editor.org/cgi-bin/rfcdoctype.pl?loc=RFC&letsgo=4443&type=ftp&file_format=txt">RFC 4443</a>
	 * @see <a href=ftp://ftp.rfc-editor.org/in-notes/rfc4861.txt">RFC 4861</a>
	 * @since 2.0.0
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 */
	interface ICMPv6 extends IP.ICMP
	{
		/**
		 * ICMPv6 field offsets.
		 */
		int	ICMPv6_TYPE_POS = 0;
		int	ICMPv6_CODE_POS = ICMPv6_TYPE_POS+2;
		int	ICMPv6_CSUM_POS = ICMPv6_CODE_POS+1;
		int	ICMPv6_HEADER_LEN = ICMPv6_CSUM_POS+2;
		
		/**
		 * ICMPv6 message types.
		 */
		int
			// RFC 4443
			// ICMPv6 error messages:
			ICMPv6_DESTINATION_UNREACHABLE = 1,	//    Destination Unreachable      (see Section 3.1)
			ICMPv6_PACKET_TOO_BIG = 2,			//    Packet Too Big               (see Section 3.2)
			ICMPv6_TIME_EXCEEDED = 3,			//    Time Exceeded                (see Section 3.3)
			ICMPv6_PARAMETER_PROBLEM = 4,		//    Parameter Problem            (see Section 3.4)
			
			//      100  Private experimentation
			//      101  Private experimentation
			//
			//      127  Reserved for expansion of ICMPv6 error messages
			
			// ICMPv6 informational messages:
			
			ICMPv6_ECHO_REQUEST =  128,			//  Echo Request                 (see Section 4.1)
			ICMPv6_ECHO_REPLY = 129,			//  Echo Reply                   (see Section 4.2)

			// RFC 3810 #5.1
			MLD_QUERY_MESSAGE	 = 130,
			// [RFC2710]
			MLDv1_LISTENER_REPORT = 131,
			// [RFC2710]			
			MLDv1_LISTENER_DONE = 132,
			// RFC 4861
			ICMPv6_ROUTER_SOLICITATION = 133,
			ICMPv6_ROUTER_ADVERTISEMENT = 134,
			ICMPv6_NEIGHBOR_SOLICITATION = 135,
			ICMPv6_NEIGHBOR_ADVERTISEMENT = 136,
			ICMPv6_REDIRECT = 137,
			// RFC 3810 #5.2
			MLD_REPORT_MESSAGE	= 143
			
			//      200  Private experimentation
			//      201  Private experimentation
			//
			//      255  Reserved for expansion of ICMPv6 informational messages
			;
		
		
		/**
		 * ICMPv6 packet class.
		 *
		 * @see <a href="http://www.rfc-editor.org/cgi-bin/rfcdoctype.pl?loc=RFC&letsgo=4443&type=ftp&file_format=txt">RFC 4443</a>
		 * @see <a href=ftp://ftp.rfc-editor.org/in-notes/rfc4861.txt">RFC 4861</a>
		 * @since 2.0.0
		 *
		 * @author Esmond Pitt
		 * @version $Revision: 1.1 $
		 */
		class Packet extends IPv6.Packet implements ICMPv6
		{
			static final long serialVersionUID = -3122794746471768498L;

			private static final Map<Integer, String>	messages = new HashMap<Integer, String>();

			static
			{
				messages.put(ICMPv6_DESTINATION_UNREACHABLE, "ICMPv6_DESTINATION_UNREACHABLE");
				messages.put(ICMPv6_PACKET_TOO_BIG, "ICMPv6_PACKET_TOO_BIG");
				messages.put(ICMPv6_TIME_EXCEEDED, "ICMPv6_TIME_EXCEEDED");
				messages.put(ICMPv6_PARAMETER_PROBLEM, "ICMPv6_DESTINATION_UNREACHABLE");

				messages.put(ICMPv6_ECHO_REQUEST, "ICMPv6_ECHO_REQUEST");
				messages.put(ICMPv6_ECHO_REPLY, "ICMPv6_ECHO_REPLY");

				messages.put(MLD_QUERY_MESSAGE, "MLD_QUERY_MESSAGE");
				messages.put(MLDv1_LISTENER_REPORT, "MLDv1_LISTENER_REPORT");
				messages.put(MLDv1_LISTENER_DONE, "MLDv1_LISTENER_DONE");
				messages.put(MLD_REPORT_MESSAGE, "MLD_REPORT_MESSAGE");

				messages.put(ICMPv6_ROUTER_SOLICITATION, "ICMPv6_ROUTER_SOLICITATION");
				messages.put(ICMPv6_ROUTER_ADVERTISEMENT, "ICMPv6_ROUTER_ADVERTISEMENT");
				messages.put(ICMPv6_NEIGHBOR_SOLICITATION, "ICMPv6_NEIGHBOR_SOLICITATION");
				messages.put(ICMPv6_NEIGHBOR_ADVERTISEMENT, "ICMPv6_NEIGHBOR_ADVERTISEMENT");
				messages.put(ICMPv6_REDIRECT, "ICMPv6_REDIRECT");
			};

			/** Creates a new instance of IPv6.ICMP.Packet */
			public Packet(int lLen, ByteBuffer bb, Timeval tv)
			{
				super(lLen, bb, tv);
			}
			
			public int	getMessageCode()
			{
				return _byteBuffer.getShort(_ethOffset+ICMPv6_TYPE_POS) & 0xffff;
			}
			
			public void	setMessageCode(short code)
			{
				_byteBuffer.putShort(_ethOffset+ICMPv6_TYPE_POS, code);
			}

			public int	getMessageMajorCode()
			{
				return _byteBuffer.get(_ethOffset+ICMPv6_TYPE_POS);
			}

			public int	getMessageMinorCode()
			{
				return _byteBuffer.get(_ethOffset+ICMPv6_CODE_POS);
			}

			public int	getICMPChecksum()
			{
				return _byteBuffer.getShort(_ethOffset+ICMPv6_CSUM_POS) & 0xffff;
			}

			@Override public byte[]	getHeader()
			{
				return getICMPHeader();
			}

			public byte[]	getICMPHeader()
			{
				byte[]	icmpHeader = new byte[ICMPv6_HEADER_LEN];
				_byteBuffer.position(_ethOffset);
				_byteBuffer.get(icmpHeader);
				return icmpHeader;
//				return PacketEncoding.extractHeader(_ethOffset, ICMPv6_HEADER_LEN, _bytes);
			}

			@Override public byte[]	getData()
			{
				return getICMPData();
			}

			public byte[]	getICMPData()
			{
				byte[]	icmpData = new byte[_byteBuffer.limit()-ICMPv6_HEADER_LEN-_ethOffset];
				_byteBuffer.position(_ethOffset+ICMPv6_HEADER_LEN);
				_byteBuffer.get(icmpData);
				return icmpData;
//				return PacketEncoding.extractData(_ethOffset, ICMPv6_HEADER_LEN, _bytes);
			}

			/**
			 * {@inheritDoc}
			 */
			public void	setChecksums()
			{
				short	checksum = (short)computeSenderChecksum(_ipOffset, ICMPv6_HEADER_LEN+getICMPData().length, ICMPv6_CSUM_POS);
				_byteBuffer.putShort(_ipOffset+ICMPv6_CSUM_POS, checksum);
				super.setChecksums();
			}
			
			/**
			 * {@inheritDoc}
			 */
			public String toColoredString(boolean colored)
			{
				StringBuffer buffer = new StringBuffer();
				buffer.append('[');
				if(colored) buffer.append(getColor());
				buffer.append("ICMPv6Packet");
				if(colored) buffer.append(AnsiEscapeSequences.RESET);
				buffer.append(": ");
				buffer.append(messages.get(getMessageCode()));
				buffer.append(", ");
				buffer.append(getSourceInetAddress().getHostAddress() + "->" +
					getDestinationInetAddress().getHostAddress());
				buffer.append(" l=" + ICMPv6_HEADER_LEN + "," +
					(_byteBuffer.limit()-_ipOffset-ICMPv6_HEADER_LEN));
				buffer.append(']');
				
				return buffer.toString();
			}
			
			/**
			 * {@inheritDoc}
			 */
			public String getColor()
			{
				return AnsiEscapeSequences.LIGHT_BLUE;
			}

		}
		
		class PacketFactory
		{
			/**
			 * Create an IPv6.ICMP packet.
			 * @param lLen Link-layer header length
			 * @param bb ByteBuffer containing packet data
			 * @param tv Timeval of packet
			 * @return an IPv6.ICMP Packet of the appropriate class: MLD.Packet, or ICMP.Packet by default.
			 */
			public static Packet	dataToPacket(int lLen, ByteBuffer bb, Timeval tv)
			{
				return new Packet(lLen, bb, tv);
			}
		}
	}
	
	/**
	 * IPv6 MLD (Multicast Listener Discovery) protocol. This replaces IGMP in IPv6.
	 *<b>Notes and quotes:</b>
	 *<ol>
	 *<li>#1 Introduction: 'MLDv2 is a translation of the IGMPv3 protocol [RFC3376] for IPv6 semantics.'
	 *<li>#5 Message Formats: 'MLDv2 is a sub-protocol of ICMPv6, that is, MLDv2 message types are a
	 * subset of ICMPv6 messages, and MLDv2 messages are identified in IPv6
	 * packets by a preceding Next Header value of 58.'
	 *</ol>
	 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/pdfrfc/rfc3810.txt.pdf">RFC 3810</a>
	 * @see <a href="ftp://ftp.rfc-editor.org/in-notes/rfc4443.txt">RFC 4443</a>
	 * @see <a href=ftp://ftp.rfc-editor.org/in-notes/rfc4861.txt">RFC 4861</a>
	 * @since 2.0.0
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 */
	interface MLD extends ICMPv6
	{
		/**
		 * MLD Query interface
		 * This is valid for ICMP type {@link ICMP#MLD_QUERY_MESSAGE}.
		 * @since 2.0.0
		 *
		 * @author Esmond Pitt
		 * @version $Revision: 1.1 $
		 */
		interface Query
		{
			// For query messages
			int	MLD_MAXIMUM_RESPONSE_CODE_POS = 4;
			int	MLD_MULTICAST_ADDRESS_POS = 8;
			int	MLD_FLAGS_POS = 12;
			int	MLD_QQIC_POS = 13;
			int MLD_SRCADDR_COUNT_POS = 14;
			int MLD_SRCADDR_POS	= 16;
			int	MLD_SFLAG_MASK = 0x0080;
			int	MLD_QRV_MASK = 0x0007;
		
			/**
			 * Get the MLD max response code.
			 * @return MLD max response code, in milliseconds < 32768, or encoded as per RFC 3810 #5.1.
			 */
			int getMaxResponseCode();
			/**
			 * Set the MLD max response code.
			 * This is only valid for type {@link ICMP#MLD_QUERY_MESSAGE}.
			 * @param code MLD max response code, in milliseconds < 32768, or encoded as per RFC 3810 #5.1.
			 */
			void	setMaxResponseCode(short code);
			/**
			 * Fetch the 'Suppress Router-Side Processing' flag (RFC 3810 #5.1.7).
			 * @return the S flag.
			 */
			boolean	getSFlag();
			/**
			 * Fetch the 'Suppress Router-Side Processing' flag (RFC 3810 #5.1.7).
			 * @param s the S flag.
			 */
			void	setSFlag(boolean s);
			/**
			 * Fetch the 'Queriers Robustness Variable' (RFC 3810 #5.1.8).
			 * @return the QRV value in the low-order 3 bits.
			 */
			int	getQRV();
			/**
			 * Fetch the 'Queriers Robustness Variable' (RFC 3810 #5.1.8).
			 * @param qrv QRV value (low-order 3 bits only).
			 */
			void	setQRV(byte qrv);
			/**
			 * Fetch the Queriers Query Interval Code (RFC 3810 #5.1.9).
			 * @return the QQIC, 8 bits.
			 */
			int	getQQIC();
			/**
			 * Set the Queriers Query Interval Code (RFC 3810 #5.1.9).
			 * @param qqic the QQIC, 8 bits.
			 */
			void	setQQIC(byte qqic);
			/**
			 * Fetch the MLD group address for a query message {@link ICMP#MLD_QUERY_MESSAGE}.
			 * @return the MLD group address.
			 * @since 2.0.0
			 */
			InetAddress getGroupInetAddress();
			/**
			 * Set the MLD group address for a query {@link ICMP#MLD_QUERY_MESSAGE}.
			 * <ul>
			 * <li>For a 'general' query, this field is all zeros and there are zero query source addresses.
			 * <li>For a 'Multicast Address Specific Query', this field is set to the m/c address of
			 * interest, and there are zero query source addresses.
			 * <li>For a 'Multicast Address and Source Specific Query', all fields are set.
			 * </ul>
			 * @param address Group address.
			 * @since 2.0.0.
			 */
			void	setGroupInetAddress(InetAddress address);
			/**
			 * Fetch the MLD source addresses for a query message {@link ICMP#MLD_QUERY_MESSAGE}.
			 * @return the MLD source addresses.
			 * @since 2.0.0
			 */
			InetAddress[] getSourceInetAddresses();
			/**
			 * Set the MLD source addresses for a query {@link ICMP#MLD_QUERY_MESSAGE}.
			 * @param address source addresses.
			 * @since 2.0.0.
			 */
			void	setSourceInetAddresses(InetAddress[] address);
		
			class Packet extends IPv6.ICMPv6.Packet implements MLD.Query, IGMPFields
			{
				static final long serialVersionUID = -7032507556476911454L;

				public Packet(int lLen, ByteBuffer bb, Timeval tv)
				{
					super(lLen, bb, tv);
				}

				public int getMaxResponseCode()
				{
					return _byteBuffer.getShort(_ipOffset+MLD_MAXIMUM_RESPONSE_CODE_POS) & 0xffff;
				}

				public void setMaxResponseCode(short code)
				{
					_byteBuffer.putShort(_ipOffset+MLD_MAXIMUM_RESPONSE_CODE_POS, code);
				}

				public boolean	getSFlag()
				{
					return (_byteBuffer.getShort(_ipOffset+MLD_FLAGS_POS) & MLD_SFLAG_MASK) != 0;
				}

				public void	setSFlag(boolean s)
				{
					short value = _byteBuffer.getShort(_ipOffset+MLD_FLAGS_POS);
					if (s)
						value |= MLD_SFLAG_MASK;
					else
						value &= ~MLD_SFLAG_MASK;
					_byteBuffer.putShort(_ipOffset+MLD_FLAGS_POS, value);
				}

				public int	getQRV()
				{
					return _byteBuffer.getShort(_ipOffset+MLD_FLAGS_POS) & MLD_QRV_MASK;
				}

				public void	setQRV(byte flags)
				{
					flags &= MLD_QRV_MASK;
					short value = _byteBuffer.getShort(_ipOffset+MLD_FLAGS_POS);
					value &= ~MLD_QRV_MASK;
					value |= flags;
					_byteBuffer.putShort(_ipOffset+MLD_FLAGS_POS, value);
				}

				public int	getQQIC()
				{
					return _byteBuffer.get(_ipOffset+MLD_QQIC_POS) & 0xff;
				}

				public void	setQQIC(byte qqic)
				{
					_byteBuffer.put(_ipOffset+MLD_QQIC_POS, qqic);
				}

				public InetAddress getGroupInetAddress()
				{
					byte[]	bytes = new byte[ADDRESS_WIDTH];
					_byteBuffer.position(_ipOffset+MLD_MULTICAST_ADDRESS_POS);
					_byteBuffer.get(bytes);
					try
					{
						return InetAddress.getByAddress(bytes);
					}
					catch (UnknownHostException exc)
					{
						// Unexpected.
						exc.printStackTrace();
						return null;
					}
				}

				public void setGroupInetAddress(InetAddress address)
				{
					_byteBuffer.position(_ipOffset+MLD_MULTICAST_ADDRESS_POS);
					_byteBuffer.put(address.getAddress());
				}

				public InetAddress[]	getSourceInetAddresses()
				{
					int	count = _byteBuffer.getShort(_ipOffset+MLD_SRCADDR_COUNT_POS) & 0xffff;
					InetAddress[]	addresses = new InetAddress[count];
					byte[]	address = new byte[ADDRESS_WIDTH];
					for (int i = 0; i < count; i++)
					{
						_byteBuffer.position(_ipOffset+MLD_SRCADDR_POS+ADDRESS_WIDTH*i);
						_byteBuffer.get(address);
						try
						{
							addresses[i] = InetAddress.getByAddress(address);
						}
						catch (UnknownHostException exc)
						{
							// unexpected
							exc.printStackTrace();
						}
					}
					return addresses;
				}

				public void	setSourceInetAddresses(InetAddress[] addresses)
				{
					// TODO assert or fix packet length
					_byteBuffer.putShort(_ipOffset+MLD_SRCADDR_COUNT_POS, (short)addresses.length);
					for (int i = 0; i < addresses.length; i++)
					{
						_byteBuffer.position(_ipOffset+MLD_SRCADDR_POS+ADDRESS_WIDTH*i);
						_byteBuffer.put(addresses[i].getAddress());
					}
				}
			}
		}
		
		/**
		 * MLD Query interface.
		 * This is valid for ICMP type {@link ICMP#MLD_REPORT_MESSAGE}.
		 * This interface is read-only by design.
		 * @since 2.0.0
		 *
		 * @author Esmond Pitt
		 * @version $Revision: 1.1 $
		 */
		interface Response
		{
			int	MULTICAST_ADDRESS_RECORD_COUNT_POS	= 6,
				MULTICAST_ADDRESS_RECORD_POS		= 8;
			
			/**
			 * Fetch the number of MulticastAddressRecords contained in this reponse.
			 * @return the number of MulticastAddressRecords contained in this reponse.
			 */
			int	getMulticastAddressRecordCount();
			/**
			 * Fetch the MulticastAddressRecords contained in this reponse.
			 * @return the MulticastAddressRecords contained in this reponse.
			 */
			Packet.MulticastAddressRecord[]	getMulticastAddressRecords();

			/**
			 * MLD packet class.
			 * @since 2.0.0
			 *
			 * @author Esmond Pitt
			 * @version $Revision: 1.1 $
			 */
			class Packet extends IPv6.ICMPv6.Packet implements MLD.Response
			{
				static final long serialVersionUID = -6721135750157905565L;

				/**
				 * MulticastAddressRecord types.
				 */
				public enum MulticastAddressRecordType
				{
					UNUSED,
					MODE_IS_INCLUDE,
					MODE_IS_EXCLUDE,
					CHANGE_TO_INCLUDE_MODE,
					CHANGE_TO_EXCLUDE_MODE,
					ALLOW_NEW_SOURCES,
					BLOCK_OLD_SOURCES
				}

				/**
				 * Multicast Address Record, for {@link ICMP#MLD_REPORT_MESSAGE}
				 * @since 2.0.0
				 *
				 * @author Esmond Pitt
				 * @version $Revision: 1.1 $
				 */
				public class MulticastAddressRecord
				{
					/**
					 * Field offsets.
					 */
					static final int	RECORD_TYPE_POS = 0,
										AUX_DATALEN_POS = 1,
										NUMBER_OF_SOURCES_POS = 2,
										MULTICAST_ADDRESS_POS = 4,
										SOURCE_ADDRESS_POS = 20;

					/**
					 * Offset of this MAR within the EthernetPacket.
					 */
					int	offset;

					MulticastAddressRecord(int offset)
					{
						this.offset = offset;
					}

					/**
					 * Fetch the record type.
					 * @return the record type.
					 */
					public MulticastAddressRecordType	getRecordType()
					{
						return  MulticastAddressRecordType.values()[_byteBuffer.get(offset+RECORD_TYPE_POS) & 0xff];
					}

					/**
					 * Fetch the auxiliary data length.
					 * @return the auxiliary data length.
					 */
					public int	getAuxDataLength()
					{
						return _byteBuffer.get(offset+AUX_DATALEN_POS) & 0xff;
					}

					/**
					 * Fetch the number of source addresses.
					 * @return the number of source addresses.
					 */
					public int	getSourceCount()
					{
						return _byteBuffer.getShort(offset+NUMBER_OF_SOURCES_POS) & 0xffff;
					}

					/**
					 * @return the multicast address.
					 */
					public InetAddress	getMulticastAddress()
					{
						byte[]	address = new byte[ADDRESS_WIDTH];
						_byteBuffer.position(offset+MULTICAST_ADDRESS_POS);
						_byteBuffer.get(address);
						try
						{
							return InetAddress.getByAddress(address);
						}
						catch (UnknownHostException exc)
						{
							// unexpected
							exc.printStackTrace();
							return null;
						}
					}

					/**
					 * Fetch the source addresses.
					 * @return the source addresses.
					 */
					public InetAddress[]	getSourceAddresses()
					{
						InetAddress[]	addresses = new InetAddress[getSourceCount()];
						byte[]	address = new byte[ADDRESS_WIDTH];
						_byteBuffer.position(offset+SOURCE_ADDRESS_POS);
						for (int i = 0; i < addresses.length; i++)
						{
							_byteBuffer.get(address);
							try
							{
								addresses[i] = InetAddress.getByAddress(address);
							}
							catch (UnknownHostException exc)
							{
								// unexpected
								exc.printStackTrace();
							}
						}
						return addresses;
					}

					/**
					 * Fetch the auxiliary data.
					 * @return the auxiliary data.
					 */
					public byte[]	getAuxiliaryData()
					{
						int	size = getAuxDataLength();
						byte[]	data = new byte[size];
						_byteBuffer.position(offset+SOURCE_ADDRESS_POS+ADDRESS_WIDTH*getSourceCount());
						_byteBuffer.get(data);
						return data;
					}
				}
				
				public Packet(int lLen, ByteBuffer bb, Timeval tv)
				{
					super(lLen, bb, tv);
				}
				
				public int	getMulticastAddressRecordCount()
				{
					return _byteBuffer.getShort(_ipOffset+MULTICAST_ADDRESS_RECORD_COUNT_POS) & 0xffff;
				}
				
				public MulticastAddressRecord[]	getMulticastAddressRecords()
				{
					MulticastAddressRecord[]	records = new MulticastAddressRecord[getMulticastAddressRecordCount()];
					for (int i = 0; i < records.length; i++)
					{
						int	offset = _ipOffset+MULTICAST_ADDRESS_RECORD_POS+ADDRESS_WIDTH*i;
						_byteBuffer.position(offset);
						records[i] = new MulticastAddressRecord(offset);
					}
					return records;
				}
			}
		}
		
	}
	
	/**
	 * TCP/IPv6 protocol.
	 *
	 * This interface is read-only by design.
	 *
	 * @see <a href="http://www.rfc-editor.org/cgi-bin/rfcdoctype.pl?loc=RFC&letsgo=4443&type=ftp&file_format=txt">RFC 4443</a>
	 * @see <a href=ftp://ftp.rfc-editor.org/in-notes/rfc4861.txt">RFC 4861</a>
	 * @since 2.0.0
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 */
	interface TCP extends IP.TCP
	{
		class Packet extends IPv6.Packet implements TCP, TCPFields, IPFields
		{
			static final long serialVersionUID = 2515547396338965562L;

			private int	_tcpHeaderLength;
			
			public Packet(int lLen, ByteBuffer bb, Timeval tv)
			{
				super(lLen, bb, tv);
//				this.packetLength = bb.getInt(_ipOffset+IPv6.IPv6_PSEUDO_PACKET_LENGTH_POS);
				byte	nextHeader = bb.get(_ipOffset+IPv6.IPv6_PSEUDO_NEXT_HEADER_POS);
				assert nextHeader == IPv6.TCP;
				try
				{
					_tcpHeaderLength = ((_byteBuffer.getShort(_ipOffset + TCP_FLAG_POS) >> 12) & 0xf) * 4;
				}
				catch (IndexOutOfBoundsException exc)
				{
					
				}
			}

			public int getSourcePort()
			{
				return _byteBuffer.getShort(_ipOffset + TCP_SP_POS) & 0xffff;
			}
			
			public int getDestinationPort()
			{
				return _byteBuffer.getShort(_ipOffset + TCP_DP_POS) & 0xffff;
			}
			
			public long getSequenceNumber()
			{
				return _byteBuffer.getInt(_ipOffset + TCP_SEQ_POS) & 0xffffffffL;
			}
			
			public long getAcknowledgmentNumber()
			{
				return _byteBuffer.getInt(_ipOffset + TCP_ACK_POS) & 0xffffffffL;
			}
			
			public int getTCPHeaderLength()
			{
				return _tcpHeaderLength;
			}
			
			public int getPayloadDataLength()
			{
				// set data (payload) length based on info in headers (note: tcpdump
				//  can return extra junk bytes which bubble up to here
				int tmpLen = getLength()-getIPHeaderLength()-_tcpHeaderLength;
				return tmpLen < 0 ? 0 : tmpLen;
			}
			
			public int getWindowSize()
			{
				return _byteBuffer.getShort(_ipOffset+TCP_WIN_POS) & 0xffff;
			}
			
			public int getTCPChecksum()
			{
				return _byteBuffer.getShort(_ipOffset+TCP_CSUM_POS) & 0xffff;
			}
			
			public int getUrgentPointer()
			{
				return _byteBuffer.getShort(_ipOffset+TCP_URG_POS) & 0xffff;
			}
			
			private int getAllFlags()
			{
				return _byteBuffer.getShort(_ipOffset+TCP_FLAG_POS) & 0xffff;
			}
			
			public boolean isUrg()
			{
				return (getAllFlags() & TCP_URG_MASK) != 0;
			}
			
			public boolean isAck()
			{
				return (getAllFlags() & TCP_ACK_MASK) != 0;
			}
			
			public boolean isPsh()
			{
				return (getAllFlags() & TCP_PSH_MASK) != 0;
			}
			
			public boolean isRst()
			{
				return (getAllFlags() & TCP_RST_MASK) != 0;
			}
			
			public boolean isSyn()
			{
				return (getAllFlags() & TCP_SYN_MASK) != 0;
			}
			
			public boolean isFin()
			{
				return (getAllFlags() & TCP_FIN_MASK) != 0;
			}
			
			
			public byte[] getTCPHeader()
			{
				byte[]	data = new byte[getTCPHeaderLength()];
				try
				{
					_byteBuffer.position(_ipOffset);
					_byteBuffer.get(data);
				}
				catch (IllegalArgumentException exc)
				{
				}
				catch (BufferUnderflowException exc)
				{
				}
				return data;
			}

			/**
			 * {@inheritDoc}
			 */
			@Override
				public byte[] getHeader()
			{
				return getTCPHeader();
			}
			
			public byte[] getTCPData()
			{
				// set data length based on info in headers (note: tcpdump
				//  can return extra junk bytes which bubble up to here
				try
				{
					_byteBuffer.position(_ipOffset+getTCPHeaderLength());
				}
				catch (IllegalArgumentException exc)
				{
				}
				catch (BufferUnderflowException exc)
				{
				}
				byte[]	data = new byte[_byteBuffer.remaining()];
				_byteBuffer.get(data);
				return data;
			}
			
			/**
			 * {@inheritDoc}
			 */
			@Override
				public byte[] getData()
			{
				return getTCPData();
			}
			
			public boolean	isValidTCPChecksum()
			{
				return getTCPChecksum() == computeSenderChecksum(_ethOffset+IP_SRC_POS, getIPHeaderLength()-IP_SRC_POS+getTCPHeader().length+getTCPData().length, TCP_CSUM_POS);
			}
			
			/**
			 * {@inheritDoc}
			 */
			@Override
				public void	setChecksums()
			{
				// The checksum is over all the data from the IP SRC and DST addresses to the entire TCP header and data
				int checksum = computeSenderChecksum(_ethOffset+IP_SRC_POS, getIPHeaderLength()-IP_SRC_POS+getTCPHeader().length+getTCPData().length, TCP_CSUM_POS);
				_byteBuffer.putShort(_ipOffset+TCP_CSUM_POS, (short)checksum);
				super.setChecksums();
			}
			
			/**
			 * {@inheritDoc}
			 */
			public String toColoredString(boolean colored)
			{
				StringBuffer buffer = new StringBuffer();
				buffer.append('[');
				if(colored) buffer.append(getColor());
				buffer.append("TCPv6Packet");
				if(colored) buffer.append(AnsiEscapeSequences.RESET);
				buffer.append(": ");
				buffer.append(getSourceInetAddress().getHostAddress());
				buffer.append('.');
				buffer.append(IPPort.getName(getSourcePort()));
				buffer.append("->");
				buffer.append(getDestinationInetAddress().getHostAddress());
				buffer.append('.');
				buffer.append(IPPort.getName(getDestinationPort()));
				if(isUrg())
					buffer.append(";urg[0x" + Integer.toHexString(getUrgentPointer()) + "]");
				if(isAck())
					buffer.append(";ack[0x" +
						Long.toHexString(getAcknowledgmentNumber()) + "]");
				if(isPsh()) buffer.append(";psh");
				if(isRst()) buffer.append(";rst");
				if(isSyn())
					buffer.append(";syn[0x" +
						Long.toHexString(getSequenceNumber()) + "]");
				if(isFin()) buffer.append(";fin");
				buffer.append(" l=" + getTCPHeaderLength() + "," +
					getPayloadDataLength());
				buffer.append(']');
				
				return buffer.toString();
			}
			
			/**
			 * {@inheritDoc}
			 */
				public String toColoredVerboseString(boolean colored)
			{
				StringBuffer buffer = new StringBuffer();
				buffer.append('[');
				if(colored) buffer.append(getColor());
				buffer.append("TCPv6Packet");
				if(colored) buffer.append(AnsiEscapeSequences.RESET);
				buffer.append(": ");
				buffer.append("sport=" + getSourcePort() + ", ");
				buffer.append("dport=" + getDestinationPort() + ", ");
				buffer.append("seqn=0x" + Long.toHexString(getSequenceNumber()) + ", ");
				buffer.append("ackn=0x" +
					Long.toHexString(getAcknowledgmentNumber()) + ", ");
				buffer.append("hlen=" + getHeaderLength() + ", ");
				buffer.append("urg=" + isUrg() + ", ");
				buffer.append("ack=" + isAck() + ", ");
				buffer.append("psh=" + isPsh() + ", ");
				buffer.append("rst=" + isRst() + ", ");
				buffer.append("syn=" + isSyn() + ", ");
				buffer.append("fin=" + isFin() + ", ");
				buffer.append("wsize=" + getWindowSize() + ", ");
				buffer.append("sum=0x" + Integer.toHexString(getChecksum()) + ", ");
				buffer.append("uptr=0x" + Integer.toHexString(getUrgentPointer()));
				buffer.append(']');
				
				return buffer.toString();
			}
			
			/**
			 * {@inheritDoc}
			 */
			public String getColor()
			{
				return AnsiEscapeSequences.YELLOW;
			}
		}
	}
	
	/**
	 * UDP/IPv6 protocol.
	 *
	 * @see <a href="http://www.rfc-editor.org/cgi-bin/rfcdoctype.pl?loc=RFC&letsgo=4443&type=ftp&file_format=txt">RFC 4443</a>
	 * @see <a href=ftp://ftp.rfc-editor.org/in-notes/rfc4861.txt">RFC 4861</a>
	 * @since 2.0.0
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 */
	interface UDP extends IP.UDP
	{
		class Packet extends IPv6.Packet implements UDP, UDPFields, IPFields
		{
			static final long serialVersionUID = 2515547396338965562L;

			int	packetLength;
			
			public Packet(int lLen, ByteBuffer bb, Timeval tv)
			{
				super(lLen, bb, tv);
				this.packetLength = bb.getInt(_ipOffset+IPv6.IPv6_PSEUDO_PACKET_LENGTH_POS);
				byte	nextHeader = bb.get(_ipOffset+IPv6.IPv6_PSEUDO_NEXT_HEADER_POS);
				assert nextHeader == IPv6.UDP;
			}

			public int getSourcePort()
			{
				return _byteBuffer.getShort(_ipOffset+UDP_SP_POS) & 0xffff;
			}
			
			public int getDestinationPort()
			{
				return _byteBuffer.getShort(_ipOffset+UDP_DP_POS) & 0xffff;
			}
			
			public int getUDPChecksum()
			{
				return _byteBuffer.getShort(_ipOffset+UDP_CSUM_POS) & 0xffff;
			}
			
			public byte[] getUDPHeader()
			{
				byte[]	bytes = new byte[UDP_HEADER_LEN];
				_byteBuffer.position(_ipOffset);
				_byteBuffer.get(bytes);
				return bytes;
			}
			
			/**
			 * {@inheritDoc}
			 */
			@Override
				public byte[] getHeader()
			{
				return getUDPHeader();
			}
			
			public byte[] getUDPData()
			{
				_byteBuffer.position(_ipOffset+UDP_HEADER_LEN);
				byte[]	bytes = new byte[_byteBuffer.remaining()];
				_byteBuffer.get(bytes);
				return bytes;
			}
			
			/**
			 * {@inheritDoc}
			 */
			@Override
				public byte[] getData()
			{
				return getUDPData();
			}
			
			/**
			 * {@inheritDoc}
			 */
			@Override
				public boolean	isValidChecksum()
			{
				return super.isValidChecksum() && isValidUDPChecksum();
			}
			
			public boolean	isValidUDPChecksum()
			{
				return getUDPChecksum() == computeSenderChecksum(_ethOffset+IP_SRC_POS, getIPHeaderLength()-IP_SRC_POS+getUDPHeader().length+getUDPData().length, UDP_CSUM_POS);
			}
		}
	}
}
