// $Id: PacketFactory.java,v 1.2 2008/08/18 05:34:37 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Copyright (C) 2008, Esmond Pitt                                         *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.net;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.sourceforge.jpcap.util.Timeval;


/**
 * This factory constructs high-level packet objects from
 * captured data streams.
 * 
 * The {@link #createUnknownPacket} method can be overridden to allow applications
 * to provide custom Packet classes for protocols of local interest.
 *<p>
 * In addition, the <code>PacketFactories.properties</code> file can be over-ridden to add additional
 * PacketFactory implementations for new protocols.
 * See the <code>PacketFactories.properties</code> file provided for details.
 * This file is loaded from whichever of jpcap.jar or jpcap-core.jar you are using, but:
 *<ul>
 *<li> If you provide one of your own in a .jar file earlier on the classpath it will be loaded <i>instead</i>.
 *<li> If you provide one of your own in the current directory when your application runs, it will be loaded <i>as well</i>.
 *</ul>
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.2 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:37 $
 */
public class PacketFactory
{
	/**
	 * Name of the .properties file.
	 * @since 2.0.0
	 */
	public static final String	PACKET_FACTORY_PROPERTIES	= "PacketFactory.properties";
	/**
	 * .properties file key prefixes.
	 * @since 2.0.0
	 */
	public static final String	ETHERNETII_PREFIX			= "EthernetII.";
	public static final String	ETHERNET8023_PREFIX		= "Ethernet802.3.";
	public static final String	FACTORY_METHOD_NAME		= "dataToPacket";
	private static PacketFactory	instance;
	protected static Properties	properties;
	protected static Map<String, Method> factories = new HashMap<String, Method>();
	protected static Map<String, Constructor<? extends Packet>> classes = new HashMap<String, Constructor<? extends Packet>>();
	private static final Class<?>[]	parameterTypes = {int.class, ByteBuffer.class, Timeval.class};
	private static Logger	logger = Logger.getLogger(PacketFactory.class.getName());
	
	static
	{
		try
		{
			properties = new Properties();
			InputStream	is = PacketFactory.class.getResourceAsStream(PACKET_FACTORY_PROPERTIES);
			if (is == null)
				logger.log(Level.SEVERE, "Resource not found: {0}", PACKET_FACTORY_PROPERTIES);
			else
				properties.load(is);
			try
			{
				is = new FileInputStream(PACKET_FACTORY_PROPERTIES);
				properties.load(is);
			}
			catch (IOException exc)
			{
				logger.log(Level.CONFIG, "loading user packet factory configuration", exc);
			}
			@SuppressWarnings("unchecked")
			Enumeration<String> keys = (Enumeration<String>)properties.propertyNames();
			while (keys.hasMoreElements())
			{
				String	key = keys.nextElement();
				String	value = properties.getProperty(key);
				Class<?> cls = Class.forName(value);
				if (value.endsWith("Factory"))
				{
					Method	method = cls.getDeclaredMethod(FACTORY_METHOD_NAME, parameterTypes);
	//				System.out.println("protocol "+protocolString+"="+method);
					factories.put(key, method);
				}
				else	// it's a packet class.
				{
					Constructor<? extends Packet>	method = cls.asSubclass(Packet.class).getConstructor(parameterTypes);
					classes.put(key, method);
				}
			}
		}
		catch (Exception exc)
		{
			exc.printStackTrace();
		}
	}
	
	/**
	 * Set the default packet factory.
	 * @param defaultFactory PacketFactory object for unrecognized packets
	 * @since 2.0.0
	 */
	public static void	setDefault(PacketFactory defaultFactory)
	{
		instance = defaultFactory;
	}
	
	/**
	 * Convert captured packet data into an object.
	 * @param linkType Link type
	 * @param bytes Packet data
	 * @return a Packet object
	 */
	public static Packet dataToPacket(int linkType, byte [] bytes)
	{
		return dataToPacket(linkType, bytes, null);
	}
	
	/**
	 * Convert captured packet data into an object.
	 * @param linkType Link type
	 * @param bytes Packet data
	 * @param tv Timeval
	 * @return a Packet object
	 */
	public static Packet dataToPacket(int linkType, byte [] bytes, Timeval tv)
	{
		return dataToPacket(linkType, ByteBuffer.wrap(bytes), tv);
	}
	
	/**
	 * Convert captured packet data into an object.
	 * @param linkType Link type
	 * @param bb Packet data in a ByteBuffer
	 * @param tv Timeval
	 * @return a Packet object
	 * @since 2.0.0
	 */
	public static Packet dataToPacket(int linkType, ByteBuffer bb, Timeval tv)
	{
		return getInstance().createPacket(linkType, bb, tv);
	}
	
	/**
	 * Get the installed PacketFactory instance. If no default has been set, reutrns an instance of this class.
	 * @return a PacketFactory instance.
	 * @since 2.0.0
	 */
	public static synchronized PacketFactory	getInstance()
	{
		if (instance == null)
			instance = new PacketFactory();
		return instance;
	}
	
	/**
	 * Returns a Packet object for the specified linkType and data.
	 * Calls {@link #createUnknownPacket} for a packet which isn't recognized by this method;
	 * if that returns null, returns an {@link EthernetPacket}.
	 * <p>
	 * Supported packet types include:
	 * <ul>
	 * <li>TCP
	 * <li>UDP
	 * <li>ICMP
	 * <li>IGMP
	 * <li>IP
	 * <li>ARP
	 * <li>STP (Spanning Tree Protocol)
	 * <li>DTP (Dynamic Trunking Protocol)
	 * <li>CDP (Cisco Discovery Protocol)
	 * </ul>
	 *
	 * @param linkType Link type
	 * @param bb ByteBuffer
	 * @param tv Timeval, or null
	 * @return a Packet corresponding to the data received.
	 *
	 * @since 2.0.0
	 */
	protected final Packet	createPacket(int linkType, ByteBuffer bb, Timeval tv)
	{
		// record the length of the headers associated with this link layer type.
		// this length is the offset to the header embedded in the packet.
		int lLen = LinkLayer.getLinkLayerLength(linkType);
		
		// extract the protocol code for the type of header embedded in the
		// link-layer of the packet
		int offset = LinkLayer.getProtoOffset(linkType);

		/*
		 * TODO all this seems to be in the wrong place.
		 * I think we need an ISLPacket class that's derived from EthernetPacket
		 * and has an EthernetPacket as a data member ...
		 */
		// Adjust for ISL encapsulation of Ethernet frames
		byte[]	dstaddr = new byte[ISL.ISL1.length];
		bb.position(0);
		bb.get(dstaddr);
//		System.out.println("PacketFactory: dstaddr="+new BigInteger(dstaddr).toString(16));
		if (Arrays.equals(dstaddr, ISL.ISL1) || Arrays.equals(dstaddr, ISL.ISL2))
		{
			offset += ISL.ISL_HEADER_SIZE;
			lLen += ISL.ISL_HEADER_SIZE;
		}
		// END ISL STUFF
		
		int ethProtocol;
		if(offset == -1)
			// if there is no embedded protocol, assume IP?
			ethProtocol = EthernetProtocols.IP;
		else
		{
			ethProtocol = bb.getShort(offset) & 0xffff;
		}
		
		// try to recognize the ethernet type.
		Packet	p = null;
		if (ethProtocol > 0x600)
		{
			String	key = PacketFactory.ETHERNETII_PREFIX+"0x"+Integer.toHexString(ethProtocol+0x10000).substring(1);
			p = createPacket(key, lLen, bb, tv);
			if (p != null)
				return p;
//			System.out.println("PacketFactory: switching on 0x"+Integer.toHexString(ethProtocol));
		}
		switch (ethProtocol)
		{
			// ARP/RARP
			case EthernetProtocols.ARP:
			case EthernetProtocols.RARP:
				return new ARPPacket(lLen, bb, tv);
			case EthernetProtocols.IP:
				return IPv4.PacketFactory.dataToPacket(lLen, bb, tv);
			// Ignored unless configured in PacketFactory.properties
//			case EthernetProtocols.IPV6:
//				return IPv6.PacketFactory.dataToPacket(lLen, bb, tv);
			default:
				// Try Ethernet 802.3
				if (ethProtocol <= 0x600)
				{
					p = LLC.PacketFactory.dataToPacket(lLen, bb, tv);
				}
				if (p == null)
				{
					// Application hook
					p = createUnknownPacket(lLen, bb, tv);
				}
				if (p != null)
					return p;
				// ethernet level code not recognized, default to anonymous packet..
				return new EthernetPacket(lLen, bb, tv);
		}
	}
	
	/**
	 * Return a Packet object for an unknown packet format.
	 * This implementation returns null.
	 * Derived classes may override, to recognize protocols of local interest.
	 * @param lLen Length of link-layer header data to ignore
	 * @param bb Packet data in a ByteBuffer
	 * @param tv Timeval, or null
	 * @return a Packet, or null.
	 *
	 * @since 2.0.0
	 */
	protected Packet	createUnknownPacket(int lLen, ByteBuffer bb, Timeval tv)
	{
		return null;
	}
	
	/**
	 * Create a packet given a properties-file key.
	 * @param key Properties-file key, e.g. Ethernet802.3Protocol.dsapssap.0xaaaa.org.0x000c.protocol.0x2000
	 * @param lLen Link-layer l=header length
	 * @param bb Packet data in a ByteBuffer
	 * @param tv Timeval of packet
	 * @return a Packet, or null.
	 */
	protected static Packet	createPacket(String key, int lLen, ByteBuffer bb, Timeval tv)
	{
		Packet	p = null;
		try
		{
			Method	method = factories.get(key);
			Constructor<? extends Packet> ctor = classes.get(key);
//			System.out.println("PacketFactory key="+key+" method="+method+" ctor="+ctor);
			if (method != null)
			{
				// found a factory
				p = (Packet)method.invoke(null, lLen, bb, tv);
//				System.out.println(method+" returned "+p);
			}
			if (p == null)
			{
				if (ctor != null)
				{
					p = ctor.newInstance(lLen, bb, tv);
//					System.out.println(ctor+" returned "+p);
				}
			}
		}
		catch (IllegalAccessException exc)
		{
			exc.printStackTrace();
		}
		catch (InstantiationException exc)
		{
			exc.printStackTrace();
		}
		catch (InvocationTargetException exc)
		{
			exc.printStackTrace();
		}
		return p;
	}
	
	private static final String _rcsid =
		"$Id: PacketFactory.java,v 1.2 2008/08/18 05:34:37 esmondpitt Exp $";
}
