// $Id: PropertyHelper.java,v 1.3 2008/08/20 01:46:29 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Copyright (c) 2008, Esmond Pitt                                         *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.awt.Color;
import java.io.InputStream;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * Property helper utility methods.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.3 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/20 01:46:29 $
 */
public class PropertyHelper
{
	private static Logger	logger = Logger.getLogger(PropertyHelper.class.getName());
	
  /**
   * Read the specified float property.
   * <p>
   * Throws an exception if the property value isn't a floating-point number.
   *
   * @param key the name of the property
   * @return the float value of the property
   */
  public static float getFloatProperty(Properties properties, Object key) {
    String string = (String)properties.get(key);

    if(string == null)
      logger.warning("couldn't find float value for '" +  key + "'");

	// This may throw NPE
    return Float.parseFloat(string);
  }

  /**
   * Read the specified integer property.
   * <p>
   * Throws an exception if the property value isn't an integer.
   *
   * @param key the name of the property
   * @return the integer value of the property
   */
  public static int getIntProperty(Properties properties, Object key) {
    String string = (String)properties.get(key);

    if(string == null)
      logger.warning("couldn't find integer value for '" + key + "'");

    return Integer.parseInt(string);	// may throw NPE
  }

  /**
   * Convert a space delimited color tuple string to a color.
   * <p>
   * Converts a string value like "255 255 0" to a color constant,
   * in this case, yellow.
   *
   * @param key the name of the property
   * @return a Color object equivalent to the provided string contents. 
   *         Returns white if the string is null or can't be converted.
   */
  public static Color getColorProperty(Properties properties, Object key) {
    String string = (String)properties.get(key);

    if(string == null) {
      System.err.println("WARN: couldn't find color tuplet under '" + 
                         key + "'");
      return Color.white;
    }

    StringTokenizer st = new StringTokenizer(string, " ");
    Color c;
    try {
      c = new Color(Integer.parseInt(st.nextToken()),
                    Integer.parseInt(st.nextToken()),
                    Integer.parseInt(st.nextToken()));
    }
    catch(NoSuchElementException e) {
      c = Color.white;
      logger.warning("invalid color spec '" + string + 
                         "' in property file");
    }

    return c;
  }

  /**
   * Convert a dot-delimited IP address to an integer.
   * <p>
   * Converts a string value like "10.0.0.5" to an integer.
   *
   * @param key the name of the property
   * @return the integer value of the specified IP number.
   * returns zero if the IP number is not valid.
   */
  public static int getIpProperty(Properties properties, Object key) {
    String string = (String)properties.get(key);

    if(string == null) {
      logger.warning("couldn't find IP value under '" + 
                         key + "'");
      return 0;
    }

    StringTokenizer st = new StringTokenizer(string, ".");
    int address;
    try {
      address = 
        Integer.parseInt(st.nextToken()) << 24 | 
        Integer.parseInt(st.nextToken()) << 16 | 
        Integer.parseInt(st.nextToken()) <<  8 | 
        Integer.parseInt(st.nextToken());
    }
    catch(NoSuchElementException e) {
      address = 0;
      System.err.println("WARN: invalid IP spec '" + string + 
                         "' in property file");
    }

    return address;
  }

  /**
   * Read the specified boolean property.
   * Converts a property value like "true" or "1" to its boolean value. 
   * <p>
   * Returns false if the property doesn't exist or can't be converted to a 
   * boolean.
   *
   * @param key the name of the property
   * @return the property value
   */
  public static boolean getBooleanProperty(Properties properties, Object key) {
    String string = (String)properties.get(key);

    if(string == null) {
      logger.warning("couldn't find boolean value for '" + key + "'");
      return false;
    }

	string = string.toLowerCase();
    return string.equals("true") ||
       string.equals("on") ||
       string.equals("yes") ||
       string.equals("1");
  }

  /**
   * Refresh property settings from an input stream.
   * Closes the input stream after loading.
   *
   * @param name Property file name
   * @param in Input stream
   * @exception IOException I/O problem
   */
  public static Properties refresh(String name, InputStream in) 
   throws IOException {
    logger.info("loading properties from " + name);
    Properties properties = new Properties();
    properties.load(in);
	in.close();
    return properties;
  }

  /**
   * Load the specified properties file from one of the specified set of paths.
   * The file is looked for both as a class resource and as a filename on the file system.
   *
   * @param paths array of String containing target paths.
   * @param fileName the name of the property file.
   * @return a populated set of properties loaded from the first file 
   * found in the set of supplied paths: if no property file is found, 
   * returns null.
   */
  public static Properties load(String [] paths, String fileName)
  {
	  File propertiesFile = null;
	  for(String path : paths)
	  {
		try
		{
			propertiesFile = new File(path, fileName);
			// Fix for 1153560: try as resource.
			// It is convenient to try it as a resource first, as getResourceAsStream returns null.
			InputStream	in = PropertyHelper.class.getResourceAsStream(propertiesFile.getPath().replaceAll("\\\\", "/"));
			if (in == null)
			{
				// Try as file.
				in = new FileInputStream(propertiesFile);
			}
			return refresh(propertiesFile.getPath(), in);
		}
		catch(IOException exc)
		{
			logger.log(Level.SEVERE, "couldn't load properties from '" + propertiesFile + "'", exc);
		}
	}
	logger.severe("could not find file '" + fileName + "' in search paths: "+Arrays.asList(paths));
	// Fix for 1153560: don't call System.exit(1), return null.
	return null;
  }

  private static String _rcsId = 
    "$Id: PropertyHelper.java,v 1.3 2008/08/20 01:46:29 esmondpitt Exp $";
}
